/*
 * Assemble a Broadcom firmware image from individual components.
 *
 * Copyright (c) Simon Vallet 2008 <svallet@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * bcm_fw_create.c
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>


#include <lzma_unsquash.h>
#include <bcm_img.h>
#include <utils.h>

extern struct create_opts * create_opts;

/**
 * Return the size of the given file, or 0 if an error occured
 */
static inline size_t get_file_size (const char * file)
{
  struct stat st;

  if (stat(file, &st) == -1) { 
    fprintf(stderr, "%s: %s -- skipping this file.\n", file, strerror(errno));
    return 0;
  }

  return st.st_size;
}

/**
 * Copy sz bytes of the given file to the given memory location
 * Return the number of copied bytes, or -1 in case of error
 */
static ssize_t copy_file (const char * file, unsigned char * ptr, size_t sz)
{
  int ifd; unsigned char * iptr; ssize_t err = -1;

  if ((ifd=open(file, O_RDONLY)) != -1) {
    if ((iptr=mmap(NULL, sz, PROT_READ, MAP_PRIVATE, ifd, 0)) != MAP_FAILED) {
      memcpy(ptr, iptr, sz); err = sz;
      munmap(iptr, sz);
    } else {
      perror(file); 
    }
    close(ifd);
  } else {
    perror(file);
  }

  return err;
}


/* Writes a two-bytes flag */
static inline size_t write_flag (unsigned char * ptr, unsigned char flag)
{
  ptr[0] = (flag) ? 0x31 : 0x00; ptr[1] = 0x00;

  return FLAG_LEN;
}

/* Writes a long unsigned int and fills up with zeros until field_len */
static inline size_t write_luint (unsigned char * ptr, size_t addr, size_t field_len)
{
  size_t len;

  snprintf((char *)ptr, field_len, "%zu", addr);
  len = strlen((char *)ptr); memset(ptr+len, 0x00, field_len-len);

  return field_len;
}

/* Writes a string and fills up with zeros to length field_len */
static inline size_t write_string (unsigned char * ptr, const char * str, size_t field_len)
{
  size_t len;

  len = strlen(str);
  strncpy((char *)ptr, str, field_len);
  if (len < field_len) {
    memset(ptr+len, 0x00, field_len - len);
  } else {
    ptr[field_len-1] = 0x00;
  }

  return field_len;
}


/* Writes the components of the broadcom header into memory */
static ssize_t create_header (unsigned char * iptr, const struct ng_bcm_hdr * header)
{
  size_t len; unsigned char * ptr; u_int32_t hdr_chksum;

  ptr = iptr;

  snprintf((char *)ptr, TAG_VER_LEN, "%u", header->version);
  len = strlen((char *)ptr); memset(ptr+len, 0x00, TAG_VER_LEN - len);
  ptr += TAG_VER_LEN;

  memcpy(ptr, header->vendor_str_1, SIG_LEN); ptr+= SIG_LEN;
  memcpy(ptr, header->vendor_str_2, SIG_LEN_2); ptr+= SIG_LEN_2;
  memcpy(ptr, header->chip_id, CHIP_ID_LEN); ptr+= CHIP_ID_LEN;
  memcpy(ptr, header->board_id, BOARD_ID_LEN); ptr+= BOARD_ID_LEN;
  ptr += write_flag(ptr, header->is_big_endian);
  ptr += write_luint(ptr, header->img_sz, IMAGE_LEN);
  ptr += write_luint(ptr, header->cfe_addr, ADDRESS_LEN);
  ptr += write_luint(ptr, header->cfe_sz, IMAGE_LEN);
  ptr += write_luint(ptr, header->rootfs_addr, ADDRESS_LEN);
  ptr += write_luint(ptr, header->rootfs_sz, IMAGE_LEN);
  ptr += write_luint(ptr, header->kernel_addr, ADDRESS_LEN);
  ptr += write_luint(ptr, header->kernel_sz, IMAGE_LEN);
  ptr += write_flag(ptr, header->is_dual_img);
  ptr += write_flag(ptr, header->is_inactive);
  memcpy(ptr, header->reserved, RESERVED_LEN); ptr+= RESERVED_LEN;

  if (is_big_endian() == header->is_big_endian) {
    *((u_int32_t *)ptr) = header->img_chksum; ptr+=TOKEN_LEN; /* Note: we assume we are zero-filled */
    hdr_chksum = bcm_crc32(iptr, TAG_LEN-TOKEN_LEN);
    *((u_int32_t *)ptr) = hdr_chksum; ptr+=TOKEN_LEN;
  } else {
    *((u_int32_t *)ptr) = swap_bytes_32(header->img_chksum); ptr+=TOKEN_LEN; /* Note: we assume we are zero-filled */
    hdr_chksum = bcm_crc32(iptr, TAG_LEN-TOKEN_LEN);
    *((u_int32_t *)ptr) = swap_bytes_32(hdr_chksum); ptr+=TOKEN_LEN;
  }

  return (ptr-iptr); /* Should be TAG_LEN */
}

int main (int argc, char ** argv)
{
  int arg_i; char cwd[PATH_MAX]; /* Hmmm */
  struct ng_bcm_hdr bcm_hdr;
  int ofd;

  unsigned char * hdr_ptr; unsigned char * img_ptr; unsigned char * ptr;

  if ((arg_i=parse_fw_create_opts(argc, argv)) == -1) { return 1; }

  if ((ofd=open(argv[arg_i], O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {

    if (create_opts->input_dir) {
      if (getcwd(cwd, PATH_MAX)) { if (chdir(create_opts->input_dir)) { perror(create_opts->input_dir); goto cleanup; } }
      else { perror(create_opts->input_dir); goto cleanup; }
    }

    /* Defaults */
    bcm_hdr.version = strtoul(BCM_TAG_VER, NULL, 10);
    write_string(bcm_hdr.vendor_str_1, BCM_SIG_1, SIG_LEN);
    write_string(bcm_hdr.vendor_str_2, BCM_SIG_2, SIG_LEN_2);
    write_string(bcm_hdr.chip_id, create_opts->chip_id, CHIP_ID_LEN);
    write_string(bcm_hdr.board_id, create_opts->board_id, BOARD_ID_LEN);
    memset(bcm_hdr.reserved, 0x00, RESERVED_LEN);
    bcm_hdr.is_big_endian = 1;
    bcm_hdr.is_dual_img = 0;
    bcm_hdr.is_inactive = 0;
    bcm_hdr.cfe_addr = _DFLT_BOOTLDR_OFF;
    bcm_hdr.rootfs_addr = _DFLT_ROOTFS_OFF;

    /* TODO: check addresses */
    if (create_opts->include_bootldr) {
      if ((bcm_hdr.cfe_sz = get_file_size(_BOOTLDR_FILE)) == 0) {
	bcm_hdr.cfe_addr = 0; create_opts->include_bootldr = 0;
      }
    } else {
      bcm_hdr.cfe_addr = 0;
    }

    if (create_opts->include_rootfs) {
      if ((bcm_hdr.rootfs_sz = get_file_size(_ROOTFS_FILE)) == 0) {
	bcm_hdr.rootfs_addr = 0; create_opts->include_rootfs = 0;
      }
    } else {
      bcm_hdr.rootfs_addr = 0;
    }
    
    if (create_opts->include_kernel) {
      if ((bcm_hdr.kernel_sz = get_file_size(_DFLT_KIMG_FILE)) == 0) {
	bcm_hdr.kernel_addr = 0; create_opts->include_kernel = 0;
      }
    } else {
      bcm_hdr.kernel_addr = 0;
    }

    bcm_hdr.img_sz = bcm_hdr.cfe_sz + bcm_hdr.rootfs_sz + bcm_hdr.kernel_sz;

    if ((hdr_ptr=mmap(NULL, TAG_LEN, PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0)) != MAP_FAILED && \
	(img_ptr=mmap(NULL, bcm_hdr.img_sz, PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0)) != MAP_FAILED) {

      /* Build the image */
      ptr = img_ptr;
      if (create_opts->include_bootldr) { 
	fprintf(stderr, "# Writing bootloader... ");
	if ((copy_file(_BOOTLDR_FILE, ptr, bcm_hdr.cfe_sz)) < 0) {
	  fprintf(stderr, "Fatal error while building image. Exiting.\n"); goto img_cleanup;
	}
	ptr += bcm_hdr.cfe_sz;
	fprintf(stderr, "done.\n");
      }

      if (create_opts->include_rootfs) {
	fprintf(stderr, "# Writing root filesystem... ");
	if ((copy_file(_ROOTFS_FILE, ptr, bcm_hdr.rootfs_sz)) < 0) {
	  fprintf(stderr, "Fatal error while building image. Exiting.\n"); goto img_cleanup;
	}
	ptr += bcm_hdr.rootfs_sz;
	fprintf(stderr, "done.\n");
      }

      if (create_opts->include_kernel) {
	fprintf(stderr, "# Writing kernel... ");
	if ((copy_file(_DFLT_KIMG_FILE, ptr, bcm_hdr.kernel_sz)) < 0) {
	  fprintf(stderr, "Fatal error while building image. Exiting.\n"); goto img_cleanup;
	}
	bcm_hdr.kernel_addr = bcm_hdr.rootfs_addr + bcm_hdr.rootfs_sz;
	ptr += bcm_hdr.kernel_sz;
	fprintf(stderr, "done.\n");
      }

      /* Build the Broadcom header if requested and write the whole image */
      if (create_opts->include_bcm_hdr) {
	fprintf(stderr, "# Building Broadcom header... ");
	bcm_hdr.img_chksum = bcm_crc32(img_ptr, bcm_hdr.img_sz);
	bcm_hdr.rk_chksum = bcm_crc32(img_ptr+bcm_hdr.cfe_sz, bcm_hdr.rootfs_sz+bcm_hdr.kernel_sz);
	if (is_big_endian() == bcm_hdr.is_big_endian) {
	  memcpy(bcm_hdr.reserved+_RSRVD_RK_CRC_OFFSET, &bcm_hdr.rk_chksum, sizeof(u_int32_t));
	} else {
	  u_int32_t chk = swap_bytes_32(bcm_hdr.rk_chksum);
	  memcpy(bcm_hdr.reserved+_RSRVD_RK_CRC_OFFSET, &chk, sizeof(u_int32_t));
	}
	create_header(hdr_ptr, &bcm_hdr);
	fprintf(stderr, "done.\n");

	write(ofd, hdr_ptr, TAG_LEN);
      }
      write(ofd, img_ptr, bcm_hdr.img_sz);
      

    img_cleanup:
      munmap(img_ptr, bcm_hdr.img_sz);
      munmap(hdr_ptr, TAG_LEN);
    } else {
      perror("mmap()");
    }

    if (create_opts->input_dir) { chdir(cwd); }

    close(ofd);
  }
  
 cleanup:
  free(create_opts);
   
  return 0;
}
