/*
 * Extract and unpack components from a Broadcom firmware image.
 *
 * Copyright (c) Simon Vallet 2008 <svallet@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * bcm_fw_extract.c
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>

#include <lzma_unsquash.h>
#include <bcm_img.h>
#include <utils.h>

extern struct extract_opts * extract_opts;
extern struct unsquash_opts * unsquash_opts;

static void display_bcm_hdr(const struct ng_bcm_hdr * hdr)
{
  size_t off=0;

  fprintf(stderr, "[0x%.2x] Header version: %u\n", off, hdr->version); off += TAG_VER_LEN;
  if (extract_opts->be_verbose) {
    fprintf(stderr, "[0x%.2x] Vendor signature (1): ", off); hex_dump(hdr->vendor_str_1, SIG_LEN); fprintf(stderr, "\n");
    off += SIG_LEN;
    fprintf(stderr, "[0x%.2x] Vendor signature (2): ", off); hex_dump(hdr->vendor_str_2, SIG_LEN_2); fprintf(stderr, "\n");
    off += SIG_LEN_2;
    fprintf(stderr, "[0x%.2x] Chip Id.: ", off); hex_dump(hdr->chip_id, CHIP_ID_LEN); fprintf(stderr, "\n");
    off += CHIP_ID_LEN;
  } else {
    off += (SIG_LEN + SIG_LEN_2 + CHIP_ID_LEN);
  }
  fprintf(stderr, "[0x%.2x] Board Id.: %s\n", off, hdr->board_id); off += BOARD_ID_LEN;
  fprintf(stderr, "[0x%.2x] Big endian ?: %d\n", off, hdr->is_big_endian); off += FLAG_LEN;
  fprintf(stderr, "[0x%.2x] Total img size: 0x%.8zx\n", off, hdr->img_sz);  off += IMAGE_LEN;
  fprintf(stderr, "[0x%.2x] CFE address: 0x%.8lx\n", off, hdr->cfe_addr); off += ADDRESS_LEN;
  fprintf(stderr, "[0x%.2x] CFE size: 0x%.8zx\n", off, hdr->cfe_sz); off += IMAGE_LEN;
  fprintf(stderr, "[0x%.2x] Root FS address: 0x%.8lx\n", off, hdr->rootfs_addr); off += ADDRESS_LEN;
  fprintf(stderr, "[0x%.2x] Root FS size: 0x%.8zx\n", off, hdr->rootfs_sz); off += IMAGE_LEN;
  fprintf(stderr, "[0x%.2x] Kernel address: 0x%.8lx\n", off, hdr->kernel_addr); off += ADDRESS_LEN;
  fprintf(stderr, "[0x%.2x] Kernel size: 0x%.8zx\n", off, hdr->kernel_sz); off += IMAGE_LEN;
  fprintf(stderr, "[0x%.2x] Dual img ?: %d\n", off, hdr->is_dual_img); off += FLAG_LEN;
  fprintf(stderr, "[0x%.2x] Inactive img ?: %d\n", off, hdr->is_inactive); off += FLAG_LEN;
  if (extract_opts->be_verbose) {
    fprintf(stderr, "[0x%.2x] Reserved: ", off); hex_dump(hdr->reserved, RESERVED_LEN); fprintf(stderr, "\n"); off += RESERVED_LEN;
    fprintf(stderr, "\t- RK checksum (0x%.2x): 0x%.8x\n", _RSRVD_RK_CRC_OFFSET, hdr->rk_chksum);
    fprintf(stderr, "[0x%.2x] Header checksum: 0x%.8x\n", off, hdr->hdr_chksum); off += TOKEN_LEN;
    fprintf(stderr, "[0x%.2x] Image checksum: 0x%.8x\n", off, hdr->img_chksum); off += TOKEN_LEN;
  }
}


static struct ng_bcm_hdr * parse_bcm_hdr (const void * img_ptr)
{
  struct ng_bcm_hdr * hdr;
  const struct _FILE_TAG * ptr;
  u_int32_t hdr_crc; u_int32_t img_crc; u_int32_t rk_crc;

  hdr = (struct ng_bcm_hdr *)malloc(sizeof(struct ng_bcm_hdr));
  ptr = (const struct _FILE_TAG *)img_ptr;
  
  hdr->version = strtoul((char *)(ptr->tagVersion), NULL, 10);
  memcpy(hdr->vendor_str_1, ptr->signiture_1, SIG_LEN);
  memcpy(hdr->vendor_str_2, ptr->signiture_2, SIG_LEN_2);
  memcpy(hdr->chip_id, ptr->chipId, CHIP_ID_LEN);
  memcpy(hdr->board_id, ptr->boardId, BOARD_ID_LEN);
  memcpy(hdr->reserved, ptr->reserved, RESERVED_LEN);
  hdr->is_big_endian = (*ptr->bigEndian != 0x00);
  hdr->is_dual_img = (*ptr->dualImage != 0x00);
  hdr->is_inactive = (*ptr->inactiveLen != 0x00);
  hdr->cfe_addr = strtoul((char *)ptr->cfeAddress, NULL, 10);
  hdr->rootfs_addr = strtoul((char *)ptr->rootfsAddress, NULL, 10);
  hdr->kernel_addr =  strtoul((char *)ptr->kernelAddress, NULL, 10);
  hdr->img_sz = strtoul((char *)ptr->totalImageLen, NULL, 10);
  hdr->cfe_sz = strtoul((char *)ptr->cfeLen, NULL, 10);
  hdr->rootfs_sz = strtoul((char *)ptr->rootfsLen, NULL, 10);
  hdr->kernel_sz = strtoul((char *)ptr->kernelLen, NULL, 10);
  hdr->hdr_chksum = *(u_int32_t *)&ptr->tagValidationToken;
  hdr->img_chksum = *(u_int32_t *)&ptr->imageValidationToken;

  hdr->rk_chksum = *(u_int32_t *)(ptr->reserved + _RSRVD_RK_CRC_OFFSET);

  if (is_big_endian() != hdr->is_big_endian) { /* Swap CRC if necessary */
    hdr->hdr_chksum = swap_bytes_32(hdr->hdr_chksum);
    hdr->img_chksum = swap_bytes_32(hdr->img_chksum);
    hdr->rk_chksum = swap_bytes_32(hdr->rk_chksum);
  }

  hdr_crc = bcm_crc32(img_ptr, TAG_LEN-TOKEN_LEN);
  img_crc = bcm_crc32(img_ptr+TAG_LEN, hdr->img_sz);
  rk_crc = bcm_crc32(img_ptr+TAG_LEN+hdr->cfe_sz, hdr->rootfs_sz + hdr->kernel_sz);

  if (hdr->img_sz != hdr->cfe_sz + hdr->rootfs_sz + hdr->kernel_sz) {
    fprintf(stderr, "## WARNING: Image size seems incorrect, should be 0x%.8x\n", hdr->cfe_sz + hdr->rootfs_sz + hdr->kernel_sz); 
  }
  if (hdr->hdr_chksum != hdr_crc) { fprintf(stderr, "## WARNING: Header CRC is incorrect, should be 0x%.8x\n", hdr_crc); }
  if (hdr->img_chksum != img_crc) { fprintf(stderr, "## WARNING: Image CRC is incorrect, should be 0x%.8x\n", img_crc); }
  if (hdr->rk_chksum != rk_crc) { fprintf(stderr, "## WARNING: RK CRC is incorrect, should be 0x%.8x\n", rk_crc); }

  return hdr;
}


/* Returns the offset in the image, or 0 in case of error */
static off_t do_extract (const void * img_ptr, const off_t offset, const size_t size, const char * file)
{
  int ofd;

  fprintf(stderr, "(offset 0x%.8lx, sz 0x%.8x, crc 0x%.8x) ", offset, size, bcm_crc32(img_ptr+offset, size));

  if ((ofd = open(file, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {
    write(ofd, img_ptr + offset, size);
    close(ofd);
  }

  return offset + size;
}

/* Max Kernel Size */
/* TODO: get rid of this */
#define MAX_KSZ 8388608U

/*
 *
 */
int main (int argc, char ** argv)
{
  char cwd[PATH_MAX]; /* Hmmm */
  int ifd; unsigned char * ifptr; struct stat ifst; 
  struct ng_bcm_hdr * bcm_hdr; off_t offset=0;
  int arg_i;

  if ((arg_i=parse_fw_extract_opts(argc, argv)) == -1) { return 1; }

  if (stat(argv[arg_i], &ifst) == -1) { perror(argv[1]); return 1; }

  if ((ifd=open(argv[arg_i], O_RDONLY)) != -1) {
    if ((ifptr=mmap(NULL, ifst.st_size, PROT_READ, MAP_PRIVATE, ifd, 0)) != MAP_FAILED) {

      bcm_hdr = parse_bcm_hdr(ifptr);
      display_bcm_hdr(bcm_hdr);

      if (extract_opts->output_dir) {
	if (getcwd(cwd, PATH_MAX)) { if (chdir(extract_opts->output_dir)) { perror(extract_opts->output_dir); goto cleanup; } }
	else { perror(extract_opts->output_dir); goto cleanup; }
      }

      if (extract_opts->dump_header) {
	fprintf(stderr, "Dumping Broadcom header... ");
	do_extract(ifptr, offset, TAG_LEN, _HDR_FILE);
	fprintf(stderr, "done.\n");
      }
      offset += TAG_LEN;

      if (extract_opts->dump_cfe) {
	if (bcm_hdr->cfe_sz) {
	  fprintf(stderr, "Dumping CFE boot loader... ");	
	  do_extract(ifptr, offset, bcm_hdr->cfe_sz, _BOOTLDR_FILE);	
	  fprintf(stderr, "done.\n");
	} else {
	  fprintf(stderr, "No bootloader found in image. Skipping.\n");
	}
      }
      offset += bcm_hdr->cfe_sz;

      if (extract_opts->dump_rootfs) {
	if (bcm_hdr->rootfs_sz) {
	  fprintf(stderr, "Dumping root filesystem... ");
	  do_extract(ifptr, offset, bcm_hdr->rootfs_sz, _ROOTFS_FILE);
	  fprintf(stderr, "done.\n");

	  if (extract_opts->do_extract) {
	    int img_fd;
	    
	    fprintf(stderr, "Extracting root filesystem... ");
	    /* Handle unsquash defaults */
	    unsquash_opts = (struct unsquash_opts *)malloc(sizeof(struct unsquash_opts));
	    unsquash_opts->show_superblock_only = 0;
	    unsquash_opts->do_list_files = 0;
	    unsquash_opts->do_list_extended = 0;
	    unsquash_opts->do_extract = 1;
	    unsquash_opts->do_force_extract = 1;
	    unsquash_opts->do_use_regex = 0;
	    unsquash_opts->be_quiet = 1;
	    unsquash_opts->output_dir = "squashfs-root";
	    
	    /*if (unsquash_opts->is_root_process) { umask(0); }*/
	    
	    if ((img_fd = open(_ROOTFS_FILE, O_RDONLY)) != -1) {
	      
	      if (read_super(_ROOTFS_FILE, img_fd) == FALSE) { return 1; }
	      read_uids_guids(img_fd);
	      do_read_fragment_table(img_fd);
	      do_uncompress_inode_table(img_fd);
	      do_uncompress_directory_table(img_fd);
	      do_dir_scan(img_fd, unsquash_opts->output_dir, NULL);
	      close(img_fd);
	    }
	    
	    fprintf(stderr, "done.\n");
	    free(unsquash_opts);
	  }
	} else {
	  fprintf(stderr, "No root filesystem found in image. Skipping.\n");
	}
      }
      offset += bcm_hdr->rootfs_sz;

      if (extract_opts->dump_kernel) {
	if (bcm_hdr->kernel_sz) {
	  fprintf(stderr, "Dumping kernel... ");
	  do_extract(ifptr, offset, bcm_hdr->kernel_sz, _DFLT_KIMG_FILE);
	  fprintf(stderr, "done.\n");
	  
	  if (extract_opts->do_extract) {
	    int k_ifd; int k_ofd; unsigned char * kptr; struct stat kst;
	    unsigned char * kbuf; ssize_t err;
	    
	    fprintf(stderr, "Extracting kernel... ");
	    if ((k_ifd=open(_DFLT_KIMG_FILE, O_RDONLY)) != -1) {
	      
	      if (fstat(k_ifd, &kst)) { perror(_DFLT_KIMG_FILE); return 1; }
	      if ((kptr=mmap(NULL, kst.st_size, PROT_READ, MAP_PRIVATE, k_ifd, 0)) != MAP_FAILED) {
		
		kbuf = malloc(MAX_KSZ);
		err = decompress_lzma_7zapi(kptr+12, kst.st_size-12, kbuf, MAX_KSZ);
		fprintf(stderr, "(sz 0x%.8x) ", err);
		
		if (err > 0) {
		  k_ofd = open("vmlinux", O_WRONLY|O_CREAT|O_TRUNC, 0644);
		  write(k_ofd, kbuf, err);
		  close(k_ofd);
		}

		free(kbuf);
		munmap(kptr, kst.st_size);
		
	      } else {
		perror("mmap");
		return -1;
	      }
	      close(k_ifd);
	    }
	    fprintf(stderr, "done.\n");
	  }
	} else {
	  fprintf(stderr, "No kernel found in image. Skipping.\n");
	}
      }
      offset += bcm_hdr->kernel_sz;

      if (extract_opts->output_dir) { chdir(cwd); }


      cleanup:
      free(bcm_hdr);
      munmap(ifptr, ifst.st_size);
    } else {
      perror(argv[arg_i]);
    }
    close(ifd);
  }
  free(extract_opts);

  return 0;
}
