/*
 * Manipulate kernel images used with Broadcom chipsets
 *
 * Copyright (c) Simon Vallet 2008 <svallet@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * bcm_kimg.c
 */
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <elf.h>
#include <7zapi.h>
#include <bcm_img.h>
#include <utils.h>

/* Should be enough for everybody ;-) */
#define MAX_KSZ 0x16777216U

extern struct kimg_opts * kimg_opts;

/* Returns the header offset, or -1 if an error occured */
static ssize_t parse_header (const unsigned char * img_ptr, size_t img_sz)
{
  size_t offset = 0; unsigned char swap_bytes;
  u_int32_t load_addr; u_int32_t entry_point; u_int32_t lz_kernel_sz;

  /* First strip the header */
  load_addr = *(u_int32_t *)img_ptr; offset += 4;
  if (load_addr == _KIMG_DFLT_LOAD_ADDR) { swap_bytes = 0; }
  else if (load_addr == _KIMG_DFLT_LOAD_ADDR_SWAP) { swap_bytes = 1; load_addr = swap_bytes_32(load_addr); }
  else {
    fprintf(stderr, "Could not find a valid kernel image header (is your load address  0x%.8x ?)\n", _KIMG_DFLT_LOAD_ADDR);
    return -1;
  }

  entry_point = *(u_int32_t *)(img_ptr + offset); offset += 4;
  if (swap_bytes) { entry_point = swap_bytes_32(entry_point); }


  lz_kernel_sz = *(u_int32_t *)(img_ptr + offset); offset += 4;
  if (swap_bytes) { lz_kernel_sz = swap_bytes_32(lz_kernel_sz); }
  if (lz_kernel_sz + _KIMG_TAG_LEN != img_sz) {
    fprintf(stderr, "The size of the image reported in the header (%u) does not match the real image size (%zu).\n",
	    lz_kernel_sz, img_sz);
    return -1;
  }

  if (kimg_opts->be_verbose) {
    fprintf(stderr, "[0X00] Kernel load address: 0x%.8x\n", load_addr);
    fprintf(stderr, "[0X04] Kernel entry point: 0x%.8x\n", entry_point);
    fprintf(stderr, "[0X08] Compressed kernel size: %u\n", lz_kernel_sz);
  }

  return offset;
} 

/* Extracts the compressed kernel
 * Returns the offset into the image, or -1 if an error occured 
 */
static ssize_t extract_img (const unsigned char * img_ptr, size_t img_sz)
{
  unsigned char * kbuf; size_t kbuf_sz = MAX_KSZ; int ksz;
  int ofd;

  /* First write the compressed kernel file */
  if ((ofd = open(kimg_opts->lz_kernel, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {
    fprintf(stderr, "Extracting compressed kernel (sz %zu) to %s... ", img_sz, kimg_opts->lz_kernel);
    write(ofd, img_ptr, img_sz);
    close(ofd);
    fprintf(stderr, "done.\n");
  } else {
    perror(kimg_opts->lz_kernel); /* Not fatal */
  }

  /* The uncompress and write the uncompressed file */
  kbuf = malloc(kbuf_sz);
  fprintf(stderr, "Decompressing kernel... ");
  if ((ksz = decompress_lzma_7zapi((unsigned char *)img_ptr, img_sz, kbuf, kbuf_sz)) > 0) {
    fprintf(stderr, "done.\n");
    if ((ofd=open(kimg_opts->raw_kernel, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {
      fprintf(stderr, "Extracting raw kernel (sz %d) to %s... ", ksz, kimg_opts->raw_kernel);
      write(ofd, kbuf, ksz);
      close(ofd);
      fprintf(stderr, "done.\n");
    } else {
      perror(kimg_opts->raw_kernel);
      goto fail;
    }
  } else {
    fprintf(stderr, "LZMA decompression failed. Exiting.\n");
    goto fail;
  }
  free(kbuf);
  
  return img_sz;

 fail:
  return -1;
}


static ssize_t extract_kernel_image (void)
{
  ssize_t out = -1; char cwd[PATH_MAX]; /* Hmmm */
  struct stat ist; int ifd; unsigned char * iptr; size_t offset = 0;

  if (stat(kimg_opts->kernel_img, &ist) == -1) { perror(kimg_opts->kernel_img); return -1; }

  if ((ifd=open(kimg_opts->kernel_img, O_RDONLY)) != -1) {

    if ((iptr=mmap(NULL, ist.st_size, PROT_READ, MAP_PRIVATE, ifd, 0)) != MAP_FAILED) {

      if (kimg_opts->dir) {
	if (getcwd(cwd, PATH_MAX)) { if (chdir(kimg_opts->dir)) { perror(kimg_opts->dir); goto cleanup; } }
	else { perror(kimg_opts->dir); goto cleanup; }
      }

      if ((offset=parse_header(iptr, ist.st_size)) != -1) {
	out = extract_img(iptr+offset, ist.st_size-offset);
      }

      if (kimg_opts->dir) { chdir(cwd); }

    cleanup:
      munmap(iptr, ist.st_size);
    } else {
      perror(kimg_opts->kernel_img);
    }
    close(ifd);
  } else {
    perror(kimg_opts->kernel_img);
  }

  return out;
}

static ssize_t create_kernel_image (void)
{
  char cwd[PATH_MAX]; /* Hmmm */
  ssize_t out = -1; int err;
  struct stat ist; int ifd; unsigned char * iptr;
  unsigned char * kbuf; size_t kbuf_sz; int kfd;

  int elf_fd; unsigned char * elf_ptr; Elf32_Ehdr * elf_hdr;
  unsigned char target_is_be = 1;
  u_int32_t load_base = kimg_opts->load_base;
  u_int32_t entry_point; u_int32_t zk_sz;
  
  /* First compress the raw kernel and get its size */
  if (stat(kimg_opts->raw_kernel, &ist) == -1) { perror(kimg_opts->raw_kernel); return 1; }

  if ((ifd=open(kimg_opts->raw_kernel, O_RDONLY)) != -1) {
    if ((iptr=mmap(NULL, ist.st_size, PROT_READ, MAP_PRIVATE, ifd, 0)) != MAP_FAILED) {

      if (kimg_opts->dir) {
	if (getcwd(cwd, PATH_MAX)) { if (chdir(kimg_opts->dir)) { perror(kimg_opts->dir); goto cleanup; } }
	else { perror(kimg_opts->dir); goto cleanup; }
      }
      
      kbuf = malloc((kbuf_sz=ist.st_size));
      fprintf(stderr, "Compressing kernel... ");
      if ((err=compress_lzma_7zapi(iptr, ist.st_size, kbuf, &kbuf_sz, _KIMG_LZMA_ALG, 
				   _KIMG_LZMA_DICTSZ, _KIMG_LZMA_FASTBYTES)) == 1) {

	fprintf(stderr, "done.\n"); zk_sz = kbuf_sz;
	if ((kfd=open(kimg_opts->lz_kernel, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {
	  fprintf(stderr, "Writing compressed kernel (sz %zu) in %s... ", kbuf_sz, kimg_opts->lz_kernel);
	  write(kfd, kbuf, kbuf_sz);
	  close(kfd);
	  fprintf(stderr, "done.\n");
	  
	  /* Get the entry point and endianness from the ELF image */
	  fprintf(stderr, "Parsing ELF header... ");
	  if ((elf_fd=open(kimg_opts->elf_kernel, O_RDONLY)) != -1) {
	    if ((elf_ptr=mmap(NULL, sizeof(Elf32_Ehdr), PROT_READ, MAP_PRIVATE, elf_fd, 0)) != MAP_FAILED) {

	      elf_hdr = (Elf32_Ehdr *)elf_ptr;
	      if (elf_hdr->e_ident[EI_MAG0] == ELFMAG0 && elf_hdr->e_ident[EI_MAG1] == ELFMAG1 && \
		  elf_hdr->e_ident[EI_MAG2] == ELFMAG2 && elf_hdr->e_ident[EI_MAG3] == ELFMAG3) {
		fprintf(stderr, "done.\n");

		target_is_be = (elf_hdr->e_ident[EI_DATA] == ELFDATA2MSB);
		if (kimg_opts->be_verbose) {
		  fprintf(stderr, "\tHost is a %s machine.\n", (is_big_endian()) ? "big-endian" : "little-endian");
		  fprintf(stderr, "\tTarget is a %s machine.\n", (target_is_be) ? "big-endian" : "little-endian");
		}

		entry_point = (is_big_endian() == target_is_be) ? elf_hdr->e_entry : swap_bytes_32(elf_hdr->e_entry);
		if (kimg_opts->be_verbose) {
		  fprintf(stderr, "\tLoad base: 0x%.8x, Entry point: 0x%.8x\n", load_base, entry_point);
		}

		if (is_big_endian() != target_is_be) { 
		  load_base = swap_bytes_32(load_base);
		  zk_sz = swap_bytes_32(zk_sz);
		  entry_point = swap_bytes_32(entry_point);
		}


		/* Finally create the header -- TODO: -L option */
		if ((kfd=open(kimg_opts->kernel_img, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != -1) {
		  
		  fprintf(stderr, "Writing kernel image (sz %zu) in %s... ", kbuf_sz+_KIMG_TAG_LEN,  kimg_opts->kernel_img);
		  write(kfd, (unsigned char*)&load_base, 4);
		  write(kfd, (unsigned char*)&entry_point, 4);
		  write(kfd, (unsigned char*)&zk_sz, 4);
		  write(kfd, kbuf, kbuf_sz);
		  close(kfd);
		  fprintf(stderr, "done.\n");
		  out = kbuf_sz+_KIMG_TAG_LEN;

		} else {
		  perror(kimg_opts->kernel_img);
		}
		
	      } else {
		fprintf(stderr, "failed: %s doesn't seem to be an ELF file. Exiting.\n", kimg_opts->elf_kernel);
	      }
	      munmap(elf_ptr, sizeof(Elf32_Ehdr));
	    }
	    close(elf_fd);
	  } else {
	    perror(kimg_opts->elf_kernel);
	  }  
	} else {
	  perror(kimg_opts->lz_kernel);
	}
      } else {
	fprintf(stderr, "LZMA compression failed.\n");
      }

      free(kbuf);
      if (kimg_opts->dir) { chdir(cwd); }

    cleanup:
      munmap(iptr, ist.st_size);
    } else {
      perror(kimg_opts->raw_kernel);
    }
    close(ifd);
  } else {
    perror(kimg_opts->kernel_img);
  }

  

  return out;
}

int main (int argc, char ** argv)
{
  int err = -1;

  if (parse_kimg_opts(argc, argv) == -1) { return 1; }

  switch (kimg_opts->todo) {
  case _KIMG_DO_EXTRACT:
    if ((err = extract_kernel_image()) != -1) { err = 0; }
    break;
  case _KIMG_DO_CREATE:
    if ((err = create_kernel_image()) != -1) { err = 0; }
    break;
  case _KIMG_DO_NOTHING:
  default:
    fprintf(stderr, "Bad invocation, exiting ('-h' for usage).\n");
    break;
  }

  free(kimg_opts);

  return -err;
}

