/*
 * Unsquash a squashfs filesystem.  This is a highly compressed read only filesystem.
 *
 * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007
 *  Phillip Lougher <phillip@lougher.demon.co.uk>
 *
 * This file has been modified to support LZMA decompression as 
 * implemented in Broadcom chipsets -- most of the code is shamelessly 
 * borrowed from nb4-unsquash.c available at
 * svn://svn.gna.org/svn/openbox4/trunk/tools/nb4-unsquash
 *
 * Changes to this file are
 * Copyright (c) Simon Vallet <svallet@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * lzma_unsquash.c
 */
#define CONFIG_SQUASHFS_1_0_COMPATIBILITY
#define CONFIG_SQUASHFS_2_0_COMPATIBILITY

#define TRUE 1
#define FALSE 0
#define Z_OK            0
#define Z_MEM_ERROR    (-4)
#define Z_BUF_ERROR    (-5)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <utime.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <regex.h>
#include <fnmatch.h>

#include <lzma_unsquash.h>
#include <utils.h>

/* Macros */
#define CALCULATE_HASH(start)	(start & 0xffff)

/* Ops */
static struct dir * squashfs_opendir (char *, unsigned int, unsigned int);
static struct dir * squashfs_opendir_2 (char *, unsigned int, unsigned int);
static int squashfs_readdir (struct dir *, char **, unsigned int *, unsigned int *, unsigned int *);
static void squashfs_closedir (struct dir *);
static struct pathnames * init_subdir (void);
static struct pathnames *add_subdir(struct pathnames *, struct pathname *);
static void free_subdir(struct pathnames *);
static void read_fragment_table (int);
static void read_fragment_table_1 (int);
static void read_fragment_table_2 (int);
static int read_block (int, off_t, off_t *, unsigned char *);
static void read_block_list (unsigned int *, unsigned char *, int);
static void read_block_list_1 (unsigned int *, unsigned char *, int);
static struct inode * read_inode (unsigned int, unsigned int);
static struct inode * read_inode_1 (unsigned int, unsigned int);
static struct inode * read_inode_2 (unsigned int, unsigned int);
static int create_inode (int, char *, struct inode *);
static int write_file (int, long long, char *, unsigned int, unsigned int, unsigned int, unsigned int, long long, unsigned char *, unsigned int);
static int write_block(int, unsigned char *, int);
static char * read_fragment (unsigned int);
static char * read_fragment_2 (unsigned int);
static int read_bytes (int, off_t, size_t, unsigned char *);
static int add_entry (struct hash_table_entry **, int, int);
static int lookup_entry (struct hash_table_entry **, int);
static void uncompress_inode_table (int, off_t, off_t);
static void uncompress_directory_table (int, off_t, off_t);
static int print_filename (char *, struct inode *);
static int matches (struct pathnames *, char *, struct pathnames **);
static int set_attributes (char *, int, uid_t, gid_t, time_t, unsigned int);
static int read_data_block(int, long long, unsigned int, unsigned char *);
static int dir_scan (int, char *, unsigned int, unsigned int, struct pathnames *);

/* Globals -- TODO: struct */
extern struct unsquash_opts * unsquash_opts;

static squashfs_super_block sBlk;
static squashfs_operations s_ops;
static unsigned char swap_bytes = 0x00;
static unsigned int * uid_table = NULL;
static unsigned int * guid_table = NULL;
static squashfs_fragment_entry * fragment_table = NULL;
static squashfs_fragment_entry_2 * fragment_table_2 = NULL;
static unsigned char * inode_table = NULL;
static unsigned char * directory_table = NULL;
static struct hash_table_entry * inode_table_hash[65536];
static struct hash_table_entry * directory_table_hash[65536];
static unsigned char * data; static unsigned char * file_data; static char * fragment_data;
static char ** created_inode;
static int file_count;static int sym_count;
static int dev_count; static int fifo_count;
static int dir_count;

static int lseek_broken = FALSE;
static unsigned char *zero_data;
static long long hole;
static unsigned int cached_frag = SQUASHFS_INVALID_FRAG;
static int global_img_fd;


static int lookup_type[] = {
	0,
	S_IFDIR,
	S_IFREG,
	S_IFLNK,
	S_IFBLK,
	S_IFCHR,
	S_IFIFO,
	S_IFSOCK,
	S_IFDIR,
	S_IFREG
};

/* For modestr */
static struct test table[] = {
	{ S_IFMT, S_IFSOCK, 0, 's' },
	{ S_IFMT, S_IFLNK, 0, 'l' },
	{ S_IFMT, S_IFBLK, 0, 'b' },
	{ S_IFMT, S_IFDIR, 0, 'd' },
	{ S_IFMT, S_IFCHR, 0, 'c' },
	{ S_IFMT, S_IFIFO, 0, 'p' },
	{ S_IRUSR, S_IRUSR, 1, 'r' },
	{ S_IWUSR, S_IWUSR, 2, 'w' },
	{ S_IRGRP, S_IRGRP, 4, 'r' },
	{ S_IWGRP, S_IWGRP, 5, 'w' },
	{ S_IROTH, S_IROTH, 7, 'r' },
	{ S_IWOTH, S_IWOTH, 8, 'w' },
	{ S_IXUSR | S_ISUID, S_IXUSR | S_ISUID, 3, 's' },
	{ S_IXUSR | S_ISUID, S_ISUID, 3, 'S' },
	{ S_IXUSR | S_ISUID, S_IXUSR, 3, 'x' },
	{ S_IXGRP | S_ISGID, S_IXGRP | S_ISGID, 6, 's' },
	{ S_IXGRP | S_ISGID, S_ISGID, 6, 'S' },
	{ S_IXGRP | S_ISGID, S_IXGRP, 6, 'x' },
	{ S_IXOTH | S_ISVTX, S_IXOTH | S_ISVTX, 9, 't' },
	{ S_IXOTH | S_ISVTX, S_ISVTX, 9, 'T' },
	{ S_IXOTH | S_ISVTX, S_IXOTH, 9, 'x' },
	{ 0, 0, 0, 0} };


static int squashfs_readdir(struct dir *dir, char **name, unsigned int *start_block, unsigned int *offset, unsigned int *type)
{
  if (dir->cur_entry == dir->dir_count) { return FALSE; }

  *name = dir->dirs[dir->cur_entry].name;
  *start_block = dir->dirs[dir->cur_entry].start_block;
  *offset = dir->dirs[dir->cur_entry].offset;
  *type = dir->dirs[dir->cur_entry].type;
  dir->cur_entry ++;

  return TRUE;
}

static void squashfs_closedir (struct dir *dir)
{
  free(dir->dirs);
  free(dir);
}


static struct dir * squashfs_opendir (char *pathname, unsigned int block_start, unsigned int offset)
{
  squashfs_dir_header dirh;
  char buffer[sizeof(squashfs_dir_entry) + SQUASHFS_NAME_LEN + 1];
  squashfs_dir_entry *dire = (squashfs_dir_entry *) buffer;
  long long start;
  int bytes;
  struct inode *i;
  int dir_count, size;
  struct dir_ent *new_dir;
  struct dir *dir;
  
  TRACE("squashfs_opendir: inode start block %d, offset %d\n", block_start, offset);
  
  if ((i = s_ops.read_inode(block_start, offset)) == NULL) {
    ERROR("squashfs_opendir: failed to read directory inode %d\n", block_start);
    return NULL;
  }
  
  start = sBlk.directory_table_start + i->start;
  bytes = lookup_entry(directory_table_hash, start);
  
  if (bytes == -1) {
    ERROR("squashfs_opendir: directory block %d not found!\n", block_start);
    return NULL;
  }
  
  if (unsquash_opts->do_list_files) { print_filename(pathname, i); }
  
  bytes += i->offset;
  size = i->data + bytes - 3;
  
  if ((dir = malloc(sizeof(struct dir))) == NULL) {
    ERROR("squashfs_opendir: malloc failed!\n");
    return NULL;
  }
  
  dir->dir_count = 0;
  dir->cur_entry = 0;
  dir->mode = i->mode;
  dir->uid = i->uid;
  dir->guid = i->gid;
  dir->mtime = i->time;
  dir->dirs = NULL;
  
  while (bytes < size) {			
    if (swap_bytes) {
      squashfs_dir_header sdirh;
      memcpy(&sdirh, directory_table + bytes, sizeof(sdirh));
      SQUASHFS_SWAP_DIR_HEADER(&dirh, &sdirh);
    } else {
      memcpy(&dirh, directory_table + bytes, sizeof(dirh));
    }
    
    dir_count = dirh.count + 1;
    TRACE("squashfs_opendir: Read directory header @ byte position %d, %d directory entries\n", bytes, dir_count);
    bytes += sizeof(dirh);
    
    while (dir_count--) {
      if  (swap_bytes) {
	squashfs_dir_entry sdire;
	memcpy(&sdire, directory_table + bytes, sizeof(sdire));
	SQUASHFS_SWAP_DIR_ENTRY(dire, &sdire);
      } else {
	memcpy(dire, directory_table + bytes, sizeof(dire));
      }
      bytes += sizeof(*dire);
      
      memcpy(dire->name, directory_table + bytes, dire->size + 1);
      dire->name[dire->size + 1] = '\0';
      TRACE("squashfs_opendir: directory entry %s, inode %d:%d, type %d\n", dire->name, dirh.start_block, dire->offset, dire->type);
      if ((dir->dir_count % DIR_ENT_SIZE) == 0) {
	if ((new_dir = realloc(dir->dirs, (dir->dir_count + DIR_ENT_SIZE) * sizeof(struct dir_ent))) == NULL) {
	  ERROR("squashfs_opendir: realloc failed!\n");
	  free(dir->dirs);
	  free(dir);
	  return NULL;
	}
	dir->dirs = new_dir;
      }
      strcpy(dir->dirs[dir->dir_count].name, dire->name);
      dir->dirs[dir->dir_count].start_block = dirh.start_block;
      dir->dirs[dir->dir_count].offset = dire->offset;
      dir->dirs[dir->dir_count].type = dire->type;
      dir->dir_count ++;
      bytes += dire->size + 1;
    }
  }
  
  return dir;
}

struct dir * squashfs_opendir_2 (char *pathname, unsigned int block_start, unsigned int offset)
{
  squashfs_dir_header_2 dirh;
  char buffer[sizeof(squashfs_dir_entry_2) + SQUASHFS_NAME_LEN + 1];
  squashfs_dir_entry_2 *dire = (squashfs_dir_entry_2 *) buffer;
  long long start;
  int bytes;
  struct inode *i;
  int dir_count, size;
  struct dir_ent *new_dir;
  struct dir *dir;
  
  TRACE("squashfs_opendir: inode start block %d, offset %d\n", block_start, offset);
  
  if ((i = s_ops.read_inode(block_start, offset)) == NULL) {
    ERROR("squashfs_opendir: failed to read directory inode %d\n", block_start);
    return NULL;
  }
  
  start = sBlk.directory_table_start + i->start;
  bytes = lookup_entry(directory_table_hash, start);
  
  if (bytes == -1) {
    ERROR("squashfs_opendir: directory block %d not found!\n", block_start);
    return NULL;
  }
  
  if (unsquash_opts->do_list_files) { print_filename(pathname, i); }
  
  bytes += i->offset;
  size = i->data + bytes;

  if ((dir = malloc(sizeof(struct dir))) == NULL) {
    ERROR("squashfs_opendir: malloc failed!\n");
    return NULL;
  }

  dir->dir_count = 0;
  dir->cur_entry = 0;
  dir->mode = i->mode;
  dir->uid = i->uid;
  dir->guid = i->gid;
  dir->mtime = i->time;
  dir->dirs = NULL;
  
  while (bytes < size) {			
    if (swap_bytes) {
      squashfs_dir_header_2 sdirh;
      memcpy(&sdirh, directory_table + bytes, sizeof(sdirh));
      SQUASHFS_SWAP_DIR_HEADER_2(&dirh, &sdirh);
    } else {
      memcpy(&dirh, directory_table + bytes, sizeof(dirh));
    }
	
    dir_count = dirh.count + 1;
    TRACE("squashfs_opendir: Read directory header @ byte position %d, %d directory entries\n", bytes, dir_count);
    bytes += sizeof(dirh);

    while (dir_count--) {
      if (swap_bytes) {
	squashfs_dir_entry_2 sdire;
	memcpy(&sdire, directory_table + bytes, sizeof(sdire));
	SQUASHFS_SWAP_DIR_ENTRY_2(dire, &sdire);
      } else {
	memcpy(dire, directory_table + bytes, sizeof(dire));
      }
      bytes += sizeof(*dire);
      
      memcpy(dire->name, directory_table + bytes, dire->size + 1);
      dire->name[dire->size + 1] = '\0';
      TRACE("squashfs_opendir: directory entry %s, inode %d:%d, type %d\n", dire->name, dirh.start_block, dire->offset, dire->type);
      if ((dir->dir_count % DIR_ENT_SIZE) == 0) {
	if ((new_dir = realloc(dir->dirs, (dir->dir_count + DIR_ENT_SIZE) * sizeof(struct dir_ent))) == NULL) {
	  ERROR("squashfs_opendir: realloc failed!\n");
	  free(dir->dirs);
	  free(dir);
	  return NULL;
	}
	dir->dirs = new_dir;
      }
      strcpy(dir->dirs[dir->dir_count].name, dire->name);
      dir->dirs[dir->dir_count].start_block = dirh.start_block;
      dir->dirs[dir->dir_count].offset = dire->offset;
      dir->dirs[dir->dir_count].type = dire->type;
      dir->dir_count ++;
      bytes += dire->size + 1;
    }
  }
  
  return dir;
}

static struct pathnames *init_subdir (void)
{
  struct pathnames *new = malloc(sizeof(struct pathnames *));
  new->count = 0;
  return new;
}


static struct pathnames *add_subdir(struct pathnames *paths, struct pathname *path)
{
  if(paths->count % PATHS_ALLOC_SIZE == 0) {
    paths = realloc(paths, sizeof(struct pathnames *) + (paths->count + PATHS_ALLOC_SIZE) * sizeof(struct pathname *));
  }

  paths->path[paths->count++] = path;
  return paths;
}


static void free_subdir(struct pathnames *paths)
{
  free(paths);
}



void do_read_fragment_table (int img_fd)
{
  s_ops.read_fragment_table(img_fd);
}

static void read_fragment_table (int img_fd)
{
  int i; int indexes = SQUASHFS_FRAGMENT_INDEXES(sBlk.fragments);
  squashfs_fragment_index fragment_table_index[indexes];

  TRACE("read_fragment_table: %d fragments, reading %d fragment indexes from 0x%llx\n", sBlk.fragments, indexes, sBlk.fragment_table_start);

  if (sBlk.fragments == 0) { return; }

  if ((fragment_table = (squashfs_fragment_entry *)malloc(sBlk.fragments * sizeof(squashfs_fragment_entry))) == NULL) {
    EXIT_UNSQUASH("read_fragment_table: failed to allocate fragment table\n");
  }

  if (swap_bytes) {
    squashfs_fragment_index sfragment_table_index[indexes];
    
    read_bytes(img_fd, sBlk.fragment_table_start, SQUASHFS_FRAGMENT_INDEX_BYTES(sBlk.fragments), \
	       (unsigned char *)sfragment_table_index);
    SQUASHFS_SWAP_FRAGMENT_INDEXES(fragment_table_index, sfragment_table_index, indexes);
  } else {
    read_bytes(img_fd, sBlk.fragment_table_start, SQUASHFS_FRAGMENT_INDEX_BYTES(sBlk.fragments), \
	       (unsigned char *)fragment_table_index);
  }

  for (i = 0; i < indexes; i++) {
    read_block(img_fd, fragment_table_index[i], NULL, ((unsigned char *) fragment_table) + (i * SQUASHFS_METADATA_SIZE));
    TRACE("Read fragment table block %d, from 0x%llx, length %d\n", i, fragment_table_index[i], length);
  }

  if (swap_bytes) {
    squashfs_fragment_entry sfragment;
    for (i = 0; i < sBlk.fragments; i++) {
      SQUASHFS_SWAP_FRAGMENT_ENTRY((&sfragment), (&fragment_table[i]));
      memcpy((char *) &fragment_table[i], (char *) &sfragment, sizeof(squashfs_fragment_entry));
    }
  }
  return;
}


static void read_fragment_table_1 (int img_fd)
{
  return;
}

static void read_fragment_table_2 (int img_fd)
{
  int i; int indexes = SQUASHFS_FRAGMENT_INDEXES_2(sBlk.fragments);
  unsigned int fragment_table_index[indexes];
  
  TRACE("read_fragment_table: %d fragments, reading %d fragment indexes from 0x%llx\n", sBlk.fragments, indexes, sBlk.fragment_table_start);

  if (sBlk.fragments == 0) { return; }

  if ((fragment_table_2 = (squashfs_fragment_entry_2 *)malloc(sBlk.fragments * sizeof(squashfs_fragment_entry))) == NULL) {
    EXIT_UNSQUASH("read_fragment_table: failed to allocate fragment table\n");
  }

  if (swap_bytes) {
    unsigned int sfragment_table_index[indexes];

    read_bytes(img_fd, sBlk.fragment_table_start, SQUASHFS_FRAGMENT_INDEX_BYTES_2(sBlk.fragments), (unsigned char *) sfragment_table_index);
    SQUASHFS_SWAP_FRAGMENT_INDEXES_2(fragment_table_index, sfragment_table_index, indexes);
  } else {
    read_bytes(img_fd, sBlk.fragment_table_start, SQUASHFS_FRAGMENT_INDEX_BYTES_2(sBlk.fragments), (unsigned char *) fragment_table_index);
  }

  for(i = 0; i < indexes; i++) {
    read_block(img_fd, fragment_table_index[i], NULL, ((unsigned char *) fragment_table_2) + (i * SQUASHFS_METADATA_SIZE));
    TRACE("Read fragment table block %d, from 0x%llx, length %d\n", i, fragment_table_index[i], length);
  }

  if (swap_bytes) {
    squashfs_fragment_entry_2 sfragment;
    for(i = 0; i < sBlk.fragments; i++) {
      SQUASHFS_SWAP_FRAGMENT_ENTRY_2((&sfragment), (&fragment_table_2[i]));
      memcpy((char *) &fragment_table_2[i], (char *) &sfragment, sizeof(squashfs_fragment_entry_2));
    }
  }
  
  return;
}


static int read_block(int img_fd, off_t start, off_t * next, unsigned char * block)
{
  unsigned short c_byte;
  int offset = 2;
	
  if (swap_bytes) {
    if (read_bytes(img_fd, start, 2, block) == FALSE) { goto failed; }

    ((unsigned char *) &c_byte)[1] = block[0];
    ((unsigned char *) &c_byte)[0] = block[1]; 

  } else {
    if (read_bytes(img_fd, start, 2, (unsigned char *)&c_byte) == FALSE) { goto failed; }
  }
  
  TRACE("read_block: block @0x%llx, %d %s bytes\n", start, SQUASHFS_COMPRESSED_SIZE(c_byte), SQUASHFS_COMPRESSED(c_byte) ? "compressed" : "uncompressed");
  
  if (SQUASHFS_CHECK_DATA(sBlk.flags)) { offset = 3; }

  if (SQUASHFS_COMPRESSED(c_byte)) {

    unsigned char buffer[SQUASHFS_METADATA_SIZE];
    int res;
    unsigned long bytes = SQUASHFS_METADATA_SIZE;
    
    c_byte = SQUASHFS_COMPRESSED_SIZE(c_byte);
    if (read_bytes(img_fd, start + offset, c_byte, buffer) == FALSE) { goto failed; }

    res = decompress_lzma_7zapi((unsigned char *) buffer, c_byte, (unsigned char *) block, bytes);
    
    if (res < 0) {
      switch (res) {
      case Z_MEM_ERROR:
	ERROR("lzma::uncompress failed, not enough memory\n"); goto failed;
	break;
      case Z_BUF_ERROR:
	ERROR("lzma::uncompress failed, not enough room in output buffer\n"); goto failed;
	break;
      default:
	ERROR("lzma::uncompress failed, unknown error %d\n", res); goto failed;
	break;
      }
    }
    /*
    switch (res) {
    case Z_OK:
      break;
    case Z_MEM_ERROR:
      ERROR("lzma::uncompress failed, not enough memory\n"); goto failed;
      break;
    case Z_BUF_ERROR:
      ERROR("lzma::uncompress failed, not enough room in output buffer\n"); goto failed;
      break;
    default:
      ERROR("lzma::uncompress failed, unknown error %d\n", res); goto failed;
      break;
    }
    */

    if (next) { *next = start + offset + c_byte; }

    return bytes;

  } else {

    c_byte = SQUASHFS_COMPRESSED_SIZE(c_byte);
    if (read_bytes(img_fd, start + offset, c_byte, block) == FALSE) { goto failed; }
    if (next) { *next = start + offset + c_byte; }

    return c_byte;
  }
  
 failed:
  return FALSE;
}


static void read_block_list(unsigned int *block_list, unsigned char *block_ptr, int blocks)
{
  if (swap_bytes) {
    unsigned int sblock_list[blocks];
    memcpy(sblock_list, block_ptr, blocks * sizeof(unsigned int));
    SQUASHFS_SWAP_INTS(block_list, sblock_list, blocks);
  } else {
    memcpy(block_list, block_ptr, blocks * sizeof(unsigned int));
  }
}

static void read_block_list_1(unsigned int *block_list, unsigned char *block_ptr, int blocks)
{
  unsigned short block_size;
  int i;
  
  for (i=0; i < blocks; i++, block_ptr += 2) {
    if (swap_bytes) {
      unsigned short sblock_size;
      memcpy(&sblock_size, block_ptr, sizeof(unsigned short));
      SQUASHFS_SWAP_SHORTS((&block_size), &sblock_size, 1);
    } else {
      memcpy(&block_size, block_ptr, sizeof(unsigned short));
    }
    
    block_list[i] = SQUASHFS_COMPRESSED_SIZE(block_size) | (SQUASHFS_COMPRESSED(block_size) ? 0 : SQUASHFS_COMPRESSED_BIT_BLOCK);
  }
}

static struct inode * read_inode(unsigned int start_block, unsigned int offset)
{
  static squashfs_inode_header header;
  long long start = sBlk.inode_table_start + start_block;
  int bytes = lookup_entry(inode_table_hash, start);
  unsigned char * block_ptr = inode_table + bytes + offset;
  static struct inode i;
  
  if (bytes == -1) { goto error; }

  if (swap_bytes) {
    squashfs_base_inode_header sinode;
    memcpy(&sinode, block_ptr, sizeof(header.base));
    SQUASHFS_SWAP_BASE_INODE_HEADER(&header.base, &sinode, sizeof(squashfs_base_inode_header));
  } else {
    memcpy(&header.base, block_ptr, sizeof(header.base));
  }

  i.uid = (uid_t) uid_table[header.base.uid];
  i.gid = header.base.guid == SQUASHFS_GUIDS ? i.uid : (uid_t) guid_table[header.base.guid];
  i.mode = lookup_type[header.base.inode_type] | header.base.mode;
  i.type = header.base.inode_type;
  i.time = header.base.mtime;
  i.inode_number = header.base.inode_number;
  
  switch (header.base.inode_type) {
  case SQUASHFS_DIR_TYPE: {
    squashfs_dir_inode_header *inode = &header.dir;
    
    if (swap_bytes) {
      squashfs_dir_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(header.dir));
      SQUASHFS_SWAP_DIR_INODE_HEADER(&header.dir, &sinode);
    } else {
      memcpy(&header.dir, block_ptr, sizeof(header.dir));
    }

    i.data = inode->file_size;
    i.offset = inode->offset;
    i.start = inode->start_block;
    break;
  }
  case SQUASHFS_LDIR_TYPE: {
    squashfs_ldir_inode_header *inode = &header.ldir;
    
    if (swap_bytes) {
      squashfs_ldir_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(header.ldir));
      SQUASHFS_SWAP_LDIR_INODE_HEADER(&header.ldir, &sinode);
    } else {
      memcpy(&header.ldir, block_ptr, sizeof(header.ldir));
    }

    i.data = inode->file_size;
    i.offset = inode->offset;
    i.start = inode->start_block;
    break;
  }
  case SQUASHFS_FILE_TYPE: {
    squashfs_reg_inode_header *inode = &header.reg;

    if (swap_bytes) {
      squashfs_reg_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(sinode));
      SQUASHFS_SWAP_REG_INODE_HEADER(inode, &sinode);
    } else {
      memcpy(inode, block_ptr, sizeof(*inode));
    }

    i.data = inode->file_size;
    i.frag_bytes = inode->fragment == SQUASHFS_INVALID_FRAG ?
      0 : inode->file_size % sBlk.block_size;
    i.fragment = inode->fragment;
    i.offset = inode->offset;
    i.blocks = inode->fragment == SQUASHFS_INVALID_FRAG ?
      (inode->file_size + sBlk.block_size - 1) >>
      sBlk.block_log : inode->file_size >> sBlk.block_log;
    i.start = inode->start_block;
    i.block_ptr = block_ptr + sizeof(*inode);
    break;
  }	
  case SQUASHFS_LREG_TYPE: {
    squashfs_lreg_inode_header *inode = &header.lreg;
    
    if (swap_bytes) {
      squashfs_lreg_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(sinode));
      SQUASHFS_SWAP_LREG_INODE_HEADER(inode, &sinode);
    } else {
      memcpy(inode, block_ptr, sizeof(*inode));
    }

    i.data = inode->file_size;
    i.frag_bytes = inode->fragment == SQUASHFS_INVALID_FRAG ?
      0 : inode->file_size % sBlk.block_size;
    i.fragment = inode->fragment;
    i.offset = inode->offset;
    i.blocks = inode->fragment == SQUASHFS_INVALID_FRAG ?
      (inode->file_size + sBlk.block_size - 1) >>
      sBlk.block_log : inode->file_size >> sBlk.block_log;
    i.start = inode->start_block;
    i.block_ptr = block_ptr + sizeof(*inode);
    break;
  }	
  case SQUASHFS_SYMLINK_TYPE: {
    squashfs_symlink_inode_header *inodep = &header.symlink;
    
    if (swap_bytes) {
      squashfs_symlink_inode_header sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_SYMLINK_INODE_HEADER(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }

    strncpy(i.symlink, (char *)block_ptr + sizeof(squashfs_symlink_inode_header), inodep->symlink_size);
    i.symlink[inodep->symlink_size] = '\0';
    i.data = inodep->symlink_size;
    break;
  }
  case SQUASHFS_BLKDEV_TYPE:
  case SQUASHFS_CHRDEV_TYPE: {
    squashfs_dev_inode_header *inodep = &header.dev;
    
    if (swap_bytes) {
      squashfs_dev_inode_header sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_DEV_INODE_HEADER(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }
    
    i.data = inodep->rdev;
    break;
  }
  case SQUASHFS_FIFO_TYPE:
  case SQUASHFS_SOCKET_TYPE:
    i.data = 0;
    break;
  default:
    ERROR("Unknown inode type %d in read_inode!\n", header.base.inode_type);
    return NULL;
  }
  return &i;

 error:
  return NULL;
}

static struct inode * read_inode_1(unsigned int start_block, unsigned int offset)
{
  static squashfs_inode_header_1 header;
  long long start = sBlk.inode_table_start + start_block;
  int bytes = lookup_entry(inode_table_hash, start);
  unsigned char *block_ptr = inode_table + bytes + offset;
  static struct inode i;
  static int inode_number = 1;
  
  if (bytes == -1) { goto error; }

  if (swap_bytes) {
    squashfs_base_inode_header_1 sinode;
    memcpy(&sinode, block_ptr, sizeof(header.base));
    SQUASHFS_SWAP_BASE_INODE_HEADER_1(&header.base, &sinode, sizeof(squashfs_base_inode_header_1));
  } else {
    memcpy(&header.base, block_ptr, sizeof(header.base));
  }

  i.uid = (uid_t) uid_table[(header.base.inode_type - 1) / SQUASHFS_TYPES * 16 + header.base.uid];
  if (header.base.inode_type == SQUASHFS_IPC_TYPE) {
    squashfs_ipc_inode_header_1 *inodep = &header.ipc;

    if (swap_bytes) {
      squashfs_ipc_inode_header_1 sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_IPC_INODE_HEADER_1(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }

    if (inodep->type == SQUASHFS_SOCKET_TYPE) {
      i.mode = S_IFSOCK | header.base.mode;
      i.type = SQUASHFS_SOCKET_TYPE;
    } else {
      i.mode = S_IFIFO | header.base.mode;
      i.type = SQUASHFS_FIFO_TYPE;
    }
    i.uid = (uid_t) uid_table[inodep->offset * 16 + inodep->uid];
  } else {
    i.mode = lookup_type[(header.base.inode_type - 1) % SQUASHFS_TYPES + 1] | header.base.mode;
    i.type = (header.base.inode_type - 1) % SQUASHFS_TYPES + 1;
  }
  i.gid = header.base.guid == 15 ? i.uid : (uid_t) guid_table[header.base.guid];
  i.time = sBlk.mkfs_time;
  i.inode_number = inode_number ++;

  switch (i.type) {
  case SQUASHFS_DIR_TYPE: {
    squashfs_dir_inode_header_1 *inode = &header.dir;
    
    if (swap_bytes) {
      squashfs_dir_inode_header_1 sinode;
      memcpy(&sinode, block_ptr, sizeof(header.dir));
      SQUASHFS_SWAP_DIR_INODE_HEADER_1(inode, &sinode);
    } else {
      memcpy(inode, block_ptr, sizeof(header.dir));
    }

    i.data = inode->file_size;
    i.start = inode->start_block;
    i.time = inode->mtime;
    break;
  }
  case SQUASHFS_FILE_TYPE: {
    squashfs_reg_inode_header_1 *inode = &header.reg;
    
    if (swap_bytes) {
      squashfs_reg_inode_header_1 sinode;
      memcpy(&sinode, block_ptr, sizeof(sinode));
      SQUASHFS_SWAP_REG_INODE_HEADER_1(inode, &sinode);
    } else {
      memcpy(inode, block_ptr, sizeof(*inode));
    }

    i.data = inode->file_size;
    i.time = inode->mtime;
    i.blocks = (inode->file_size + sBlk.block_size - 1) >>
      sBlk.block_log;
    i.start = inode->start_block;
    i.block_ptr = block_ptr + sizeof(*inode);
    i.fragment = 0;
    i.frag_bytes = 0;
    i.offset = 0;
    break;
  }	
  case SQUASHFS_SYMLINK_TYPE: {
    squashfs_symlink_inode_header_1 *inodep = &header.symlink;
    
    if (swap_bytes) {
      squashfs_symlink_inode_header_1 sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_SYMLINK_INODE_HEADER_1(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }
    
    strncpy(i.symlink, (char *)block_ptr + sizeof(squashfs_symlink_inode_header_1), inodep->symlink_size);
    i.symlink[inodep->symlink_size] = '\0';
    i.data = inodep->symlink_size;
    break;
  }
  case SQUASHFS_BLKDEV_TYPE:
  case SQUASHFS_CHRDEV_TYPE: {
    squashfs_dev_inode_header_1 *inodep = &header.dev;
    
    if (swap_bytes) {
      squashfs_dev_inode_header_1 sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_DEV_INODE_HEADER_1(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }

    i.data = inodep->rdev;
    break;
  }
  case SQUASHFS_FIFO_TYPE:
  case SQUASHFS_SOCKET_TYPE: {
    i.data = 0;
    break;
  }
  default:
    ERROR("Unknown inode type %d in read_inode_header_1!\n", header.base.inode_type);
    return NULL;
  }
  return &i;

 error:
  return NULL;
}

static struct inode * read_inode_2(unsigned int start_block, unsigned int offset)
{
  static squashfs_inode_header_2 header;
  long long start = sBlk.inode_table_start + start_block;
  int bytes = lookup_entry(inode_table_hash, start);
  unsigned char *block_ptr = inode_table + bytes + offset;
  static struct inode i;
  static int inode_number = 1;

  if (bytes == -1) { goto error; }

  if (swap_bytes) {
    squashfs_base_inode_header_2 sinode;
    memcpy(&sinode, block_ptr, sizeof(header.base));
    SQUASHFS_SWAP_BASE_INODE_HEADER_2(&header.base, &sinode, sizeof(squashfs_base_inode_header_2));
  } else {
    memcpy(&header.base, block_ptr, sizeof(header.base));
  }

  i.uid = (uid_t) uid_table[header.base.uid];
  i.gid = header.base.guid == SQUASHFS_GUIDS ? i.uid : (uid_t) guid_table[header.base.guid];
  i.mode = lookup_type[header.base.inode_type] | header.base.mode;
  i.type = header.base.inode_type;
  i.time = sBlk.mkfs_time;
  i.inode_number = inode_number++;
  
  switch (header.base.inode_type) {
  case SQUASHFS_DIR_TYPE: {
    squashfs_dir_inode_header_2 *inode = &header.dir;
    
    if (swap_bytes) {
      squashfs_dir_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(header.dir));
      SQUASHFS_SWAP_DIR_INODE_HEADER_2(&header.dir, &sinode);
    } else {
      memcpy(&header.dir, block_ptr, sizeof(header.dir));
    }

    i.data = inode->file_size;
    i.offset = inode->offset;
    i.start = inode->start_block;
    i.time = inode->mtime;
    break;
  }
  case SQUASHFS_LDIR_TYPE: {
    squashfs_ldir_inode_header_2 *inode = &header.ldir;
    
    if (swap_bytes) {
      squashfs_ldir_inode_header sinode;
      memcpy(&sinode, block_ptr, sizeof(header.ldir));
      SQUASHFS_SWAP_LDIR_INODE_HEADER_2(&header.ldir, &sinode);
    } else {
      memcpy(&header.ldir, block_ptr, sizeof(header.ldir));
    }

    i.data = inode->file_size;
    i.offset = inode->offset;
    i.start = inode->start_block;
    i.time = inode->mtime;
    break;
  }
  case SQUASHFS_FILE_TYPE: {
    squashfs_reg_inode_header_2 *inode = &header.reg;
    
    if (swap_bytes) {
      squashfs_reg_inode_header_2 sinode;
      memcpy(&sinode, block_ptr, sizeof(sinode));
      SQUASHFS_SWAP_REG_INODE_HEADER_2(inode, &sinode);
    } else {
      memcpy(inode, block_ptr, sizeof(*inode));
    }

    i.data = inode->file_size;
    i.time = inode->mtime;
    i.frag_bytes = inode->fragment == SQUASHFS_INVALID_FRAG ?
      0 : inode->file_size % sBlk.block_size;
    i.fragment = inode->fragment;
    i.offset = inode->offset;
    i.blocks = inode->fragment == SQUASHFS_INVALID_FRAG ?
      (inode->file_size + sBlk.block_size - 1) >>
      sBlk.block_log : inode->file_size >> sBlk.block_log;
    i.start = inode->start_block;
    i.block_ptr = block_ptr + sizeof(*inode);
    break;
  }	
  case SQUASHFS_SYMLINK_TYPE: {
    squashfs_symlink_inode_header_2 *inodep = &header.symlink;
    
    if (swap_bytes) {
      squashfs_symlink_inode_header_2 sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_SYMLINK_INODE_HEADER_2(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }
    
    strncpy(i.symlink, (char *)block_ptr + sizeof(squashfs_symlink_inode_header_2), inodep->symlink_size);
    i.symlink[inodep->symlink_size] = '\0';
    i.data = inodep->symlink_size;
    break;
  }
  case SQUASHFS_BLKDEV_TYPE:
  case SQUASHFS_CHRDEV_TYPE: {
    squashfs_dev_inode_header_2 *inodep = &header.dev;
    
    if (swap_bytes) {
      squashfs_dev_inode_header_2 sinodep;
      memcpy(&sinodep, block_ptr, sizeof(sinodep));
      SQUASHFS_SWAP_DEV_INODE_HEADER_2(inodep, &sinodep);
    } else {
      memcpy(inodep, block_ptr, sizeof(*inodep));
    }

    i.data = inodep->rdev;
    break;
  }
  case SQUASHFS_FIFO_TYPE:
  case SQUASHFS_SOCKET_TYPE:
    i.data = 0;
    break;
  default:
    ERROR("Unknown inode type %d in read_inode_header_2!\n", header.base.inode_type);
    return NULL;
  }
  return &i;

 error:
  return NULL;
}


static int read_data_block (int img_fd, long long start, unsigned int size, unsigned char *block)
{
  int res;
  unsigned long bytes = sBlk.block_size;
  int c_byte = SQUASHFS_COMPRESSED_SIZE_BLOCK(size);
  
  TRACE("read_data_block: block @0x%llx, %d %s bytes\n", start, SQUASHFS_COMPRESSED_SIZE_BLOCK(c_byte), SQUASHFS_COMPRESSED_BLOCK(c_byte) ? "compressed" : "uncompressed");
  
  if (SQUASHFS_COMPRESSED_BLOCK(size)) {
    
    if (read_bytes(img_fd, start, c_byte, data) == FALSE) { return 0; }
	  
    res = decompress_lzma_7zapi((unsigned char *) data, c_byte, (unsigned char *)block, bytes);

    if (res < 0) {
      if(res == Z_MEM_ERROR) {
	ERROR("lzma::uncompress failed, not enough memory\n");
      } else if (res == Z_BUF_ERROR) {
	ERROR("lzma::uncompress failed, not enough room in output buffer\n");
      } else {
	ERROR("lzma::uncompress failed, unknown error %d\n", res);
      }
      return 0;
    }
    /*
    if (res != Z_OK) {
      if(res == Z_MEM_ERROR) {
	ERROR("lzma::uncompress failed, not enough memory\n");
      } else if (res == Z_BUF_ERROR) {
	ERROR("lzma::uncompress failed, not enough room in output buffer\n");
      } else {
	ERROR("lzma::uncompress failed, unknown error %d\n", res);
      }
      return 0;
    }
    */
	  
    return bytes;

  } else {

    if (read_bytes(img_fd, start, c_byte, block) == FALSE) { return 0; }
    
    return c_byte;
  }
}


static int create_inode (int img_fd, char *pathname, struct inode *i)
{
  TRACE("create_inode: pathname %s\n", pathname);

  if (created_inode[i->inode_number - 1]) {
    TRACE("create_inode: hard link\n");
    if (unsquash_opts->do_force_extract) { unlink(pathname); }

    if (link(created_inode[i->inode_number - 1], pathname) == -1) {
      ERROR("create_inode: failed to create hardlink, because %s\n", strerror(errno));
      return FALSE;
    }

    return TRUE;
  }

  switch (i->type) {
  case SQUASHFS_FILE_TYPE:
  case SQUASHFS_LREG_TYPE:
    TRACE("create_inode: regular file, file_size %lld, blocks %d\n", i->data, i->blocks);
    
    if (write_file(img_fd, i->data, pathname, i->fragment, i->frag_bytes,
		  i->offset, i->blocks, i->start, i->block_ptr, i->mode)) {
      set_attributes(pathname, i->mode, i->uid, i->gid, i->time, unsquash_opts->do_force_extract);
      file_count ++;
    }
    break;
  case SQUASHFS_SYMLINK_TYPE:
    TRACE("create_inode: symlink, symlink_size %d\n", i->data);
    
    if (unsquash_opts->do_force_extract) { unlink(pathname); }

    if (symlink(i->symlink, pathname) == -1) {
      ERROR("create_inode: failed to create symlink %s, because %s\n", pathname,
	    strerror(errno));
      break;
    }

    if ((geteuid() == SYSTEM_UID)) {
      if (lchown(pathname, i->uid, i->gid) == -1)
	ERROR("create_inode: failed to change uid and gids on %s, because %s\n", pathname, strerror(errno));
    }

    sym_count ++;
    break;
  case SQUASHFS_BLKDEV_TYPE:
  case SQUASHFS_CHRDEV_TYPE: {
    int chrdev = i->type == SQUASHFS_CHRDEV_TYPE;
    TRACE("create_inode: dev, rdev 0x%x\n", i->data);
    
    if ((geteuid() == SYSTEM_UID)) {
    
      if (unsquash_opts->do_force_extract) { unlink(pathname); }

      if (mknod(pathname, chrdev ? S_IFCHR : S_IFBLK,
		makedev((i->data >> 8) & 0xff, i->data & 0xff)) == -1) {
			    
	ERROR("create_inode: failed to create %s device %s, because %s\n",
	      chrdev ? "character" : "block", pathname, strerror(errno));
	break;
      }
      set_attributes(pathname, i->mode, i->uid, i->gid, i->time, TRUE);
      dev_count ++;
    
    } else {
      ERROR("create_inode: could not create %s device %s, because you're not superuser!\n",
	    chrdev ? "character" : "block", pathname);
    }
    break;
  }
  case SQUASHFS_FIFO_TYPE:
    TRACE("create_inode: fifo\n");
    
    if (unsquash_opts->do_force_extract) { unlink(pathname); }

    if (mknod(pathname, S_IFIFO, 0) == -1) {
      ERROR("create_inode: failed to create fifo %s, because %s\n",
	    pathname, strerror(errno));
      break;
    }
    set_attributes(pathname, i->mode, i->uid, i->gid, i->time, TRUE);
    fifo_count ++;
    break;
  case SQUASHFS_SOCKET_TYPE:
    TRACE("create_inode: socket\n");
    ERROR("create_inode: socket %s ignored\n", pathname);
    break;
  default:
    ERROR("Unknown inode type %d in create_inode_table!\n", i->type);
    return FALSE;
  }

  created_inode[i->inode_number - 1] = strdup(pathname);

  return TRUE;
}


static int write_block(int file_fd, unsigned char *buffer, int size)
{
  off_t off = hole;

  if (hole) {
    if (lseek_broken == FALSE && lseek(file_fd, off, SEEK_CUR) == -1) {
      /* failed to seek beyond end of file */
      if ((zero_data = malloc(sBlk.block_size)) == NULL) {
	EXIT_UNSQUASH("write_block: failed to alloc zero data block\n");
      }
      memset(zero_data, 0, sBlk.block_size);
      lseek_broken = TRUE;
    }
    
    if (lseek_broken) {
      int blocks = (hole + sBlk.block_size -1) / sBlk.block_size;
      int avail_bytes, i;
      for(i = 0; i < blocks; i++, hole -= avail_bytes) {
	avail_bytes = hole > sBlk.block_size ? sBlk.block_size : hole;
	if(write(file_fd, zero_data, avail_bytes) < avail_bytes) { goto failure; }
      }
    }
    hole = 0;
  }

  if (write(file_fd, buffer, size) < size) { goto failure; }
  
  return TRUE;

failure:
  return FALSE;
}

	
static int write_file(int img_fd, long long file_size, char *pathname, unsigned int fragment, unsigned int frag_bytes,
		      unsigned int offset, unsigned int blocks, long long start, unsigned char *block_ptr,
		      unsigned int mode)
{
  unsigned int file_fd, bytes, i;
  unsigned int *block_list;
  int file_end = file_size / sBlk.block_size;

  TRACE("write_file: regular file, blocks %d, file_size %d\n", blocks,file_size);

  hole = 0;
  if ((block_list = malloc(blocks * sizeof(unsigned int))) == NULL) {
    ERROR("write_file: unable to malloc block list\n");
    return FALSE;
  }

  s_ops.read_block_list(block_list, (unsigned char *)block_ptr, blocks);

  if ((file_fd = open(pathname, O_CREAT | O_WRONLY | (unsquash_opts->do_force_extract ? O_TRUNC : 0), (mode_t) mode & 0777)) == -1) {
    ERROR("write_file: failed to create file %s, because %s\n", pathname,
	  strerror(errno));
    free(block_list);
    return FALSE;
  }

  for (i = 0; i < blocks; i++) {
    if (block_list[i] == 0) { /* sparse file */
      hole += i == file_end ? file_size & (sBlk.block_size - 1) : sBlk.block_size;
      continue;
    }
    
    if ((bytes = read_data_block(img_fd, start, block_list[i], file_data)) == 0) {
      ERROR("write_file: failed to read data block 0x%llx\n", start);
      goto failure;
    }
    
    if (i < blocks-1) { bytes=65536; }
    else { bytes=file_size - (blocks-1)*65536 - frag_bytes; }

    TRACE("decompress %d bytes into %d bytes\n", SQUASHFS_COMPRESSED_SIZE_BLOCK(block_list[i]),bytes);

    if (write_block(file_fd, file_data, bytes) == FALSE) {
      ERROR("write_file: failed to write data block 0x%llx\n", start);
      goto failure;
    }

    start += SQUASHFS_COMPRESSED_SIZE_BLOCK(block_list[i]);
  }

  if (frag_bytes != 0) {
    char * fragment_data = s_ops.read_fragment(fragment);

    if (fragment_data == NULL) { goto failure; }

    if (write_block(file_fd, (unsigned char *)fragment_data + offset, frag_bytes) == FALSE) {
      ERROR("write_file: failed to write fragment %d\n", fragment);
      goto failure;
    }
  }

  if (hole) {
    /* corner case for hole extending to end of file */
    hole --;
    if (write_block(file_fd, 0x00, 1) == FALSE) {
      ERROR("write_file: failed to write sparse data block\n");
      goto failure;
    }
  }

  close(file_fd);
  free(block_list);
  return TRUE;

failure:
  close(file_fd);
  free(block_list);
  return FALSE;
}

static int set_attributes (char *pathname, int mode, uid_t uid, gid_t guid, time_t time, unsigned int set_mode)
{
  struct utimbuf times = { time, time };
  
  if (utime(pathname, &times) == -1) {
    ERROR("set_attributes: failed to set time on %s, because %s\n", pathname, strerror(errno));
    return FALSE;
  }

  if ((geteuid() == SYSTEM_UID)) {
    if (chown(pathname, uid, guid) == -1) {
      ERROR("set_attributes: failed to change uid and gids on %s, because %s\n", pathname, strerror(errno));
      ERROR("chown(%s, %d, %d)\n", pathname, uid, guid);
      return FALSE;
    }
  } else {
    mode &= ~07000;
  }

  if ((set_mode || (mode & 07000)) && chmod(pathname, (mode_t) mode) == -1) {
    ERROR("set_attributes: failed to change mode %s, because %s\n", pathname, strerror(errno));
    return FALSE;
  }

  return TRUE;
}

static char *read_fragment (unsigned int fragment)
{
  TRACE("read_fragment: reading fragment %d\n", fragment);
  
  if (cached_frag == SQUASHFS_INVALID_FRAG || fragment != cached_frag) {
  
    squashfs_fragment_entry *fragment_entry = &fragment_table[fragment];
    if (read_data_block(global_img_fd, fragment_entry->start_block, fragment_entry->size, (unsigned char *)fragment_data) == 0) {
      ERROR("read_fragment: failed to read fragment %d\n", fragment);
      cached_frag = SQUASHFS_INVALID_FRAG;
      return NULL;
    }
    cached_frag = fragment;
  }

  return fragment_data;
}


static char *read_fragment_2 (unsigned int fragment)
{
  TRACE("read_fragment: reading fragment %d\n", fragment);
  
  if (cached_frag == SQUASHFS_INVALID_FRAG || fragment != cached_frag) {
  
    squashfs_fragment_entry_2 *fragment_entry = &fragment_table_2[fragment];
    if (read_data_block(global_img_fd, fragment_entry->start_block, fragment_entry->size, (unsigned char *)fragment_data) == 0) {
      ERROR("read_fragment: failed to read fragment %d\n", fragment);
      cached_frag = SQUASHFS_INVALID_FRAG;
      return NULL;
    }
    cached_frag = fragment;

  }

  return fragment_data;
}


static int read_bytes(int fd, off_t offset, size_t bytes, unsigned char * buf)
{
  TRACE("read_bytes: reading from position 0x%llx, bytes %d\n", byte, bytes);

  if (lseek(fd, offset, SEEK_SET) == -1) {
    ERROR("Lseek failed because %s\b", strerror(errno));
    return FALSE;
  }

  if (read(fd, buf, bytes) == -1) {
    ERROR("Read on destination failed because %s\n", strerror(errno));
    return FALSE;
  }

  return TRUE;
}


int read_super (const char * img_path, int img_fd)
{
  global_img_fd = img_fd;
  read_bytes(img_fd, SQUASHFS_START, sizeof(squashfs_super_block), (unsigned char *)&sBlk);

  /* Check it is a SQUASHFS superblock */
  swap_bytes = 0;
  if (sBlk.s_magic != SQUASHFS_MAGIC) {
    if (sBlk.s_magic == SQUASHFS_MAGIC_SWAP) {
      squashfs_super_block sblk;
      ERROR("Reading a different endian SQUASHFS filesystem on %s\n", img_path);
      SQUASHFS_SWAP_SUPER_BLOCK(&sblk, &sBlk);
      memcpy(&sBlk, &sblk, sizeof(squashfs_super_block));
      swap_bytes = 1;
    } else  {
      ERROR("Can't find a SQUASHFS superblock on %s\n", img_path);
      return FALSE;
    }
  }

  /* Check the MAJOR & MINOR versions */
  if (sBlk.s_major == 1 || sBlk.s_major == 2) {
    sBlk.bytes_used = sBlk.bytes_used_2;
    sBlk.uid_start = sBlk.uid_start_2;
    sBlk.guid_start = sBlk.guid_start_2;
    sBlk.inode_table_start = sBlk.inode_table_start_2;
    sBlk.directory_table_start = sBlk.directory_table_start_2;
		
    if(sBlk.s_major == 1) {
      sBlk.block_size = sBlk.block_size_1;
      sBlk.fragment_table_start = sBlk.uid_start;
      s_ops.squashfs_opendir = squashfs_opendir_2;
      s_ops.read_fragment_table = read_fragment_table_1;
      s_ops.read_block_list = read_block_list_1;
      s_ops.read_inode = read_inode_1;
    } else {
      sBlk.fragment_table_start = sBlk.fragment_table_start_2;
      s_ops.squashfs_opendir = squashfs_opendir_2;
      s_ops.read_fragment = read_fragment_2;
      s_ops.read_fragment_table = read_fragment_table_2;
      s_ops.read_block_list = read_block_list;
      s_ops.read_inode = read_inode_2;
    }
  } else if (sBlk.s_major == 3 && sBlk.s_minor <= 1) {
    s_ops.squashfs_opendir = squashfs_opendir;
    s_ops.read_fragment = read_fragment;
    s_ops.read_fragment_table = read_fragment_table;
    s_ops.read_block_list = read_block_list;
    s_ops.read_inode = read_inode;
  } else {
    ERROR("Filesystem on %s is (%d:%d), ", img_path, sBlk.s_major, sBlk.s_minor);
    ERROR("which is a later filesystem version than I support!\n");
    return FALSE;
  }

  /* Initialize */
  if ((fragment_data = malloc(sBlk.block_size)) == NULL) { EXIT_UNSQUASH("failed to allocate fragment_data\n"); }

  if ((file_data = malloc(sBlk.block_size)) == NULL) { EXIT_UNSQUASH("failed to allocate file_data"); }

  if ((data = malloc(sBlk.block_size)) == NULL) { EXIT_UNSQUASH("failed to allocate datan\n"); }

  if ((created_inode = malloc(sBlk.inodes * sizeof(char *))) == NULL) { EXIT_UNSQUASH("failed to allocate created_inode\n"); }
  else { memset(created_inode, 0, sBlk.inodes * sizeof(char *)); }

  return TRUE;
}

void squashfs_stat (const char * img_path)
{
  time_t mkfs_time = (time_t) sBlk.mkfs_time;
  char *mkfs_str = ctime(&mkfs_time);

#if __BYTE_ORDER == __BIG_ENDIAN
  fprintf(stderr, "Found a valid %s endian SQUASHFS %d:%d superblock on %s.\n", swap_bytes ? "little" : "big", sBlk.s_major, sBlk.s_minor, img_path);
#else
  fprintf(stderr, "Found a valid %s endian SQUASHFS %d:%d superblock on %s.\n", swap_bytes ? "big" : "little", sBlk.s_major, sBlk.s_minor, img_path);
#endif

  fprintf(stderr, "Creation or last append time %s", mkfs_str ? mkfs_str : "failed to get time\n");
  fprintf(stderr, "Filesystem is %sexportable via NFS\n", SQUASHFS_EXPORTABLE(sBlk.flags) ? "" : "not ");

  fprintf(stderr, "Inodes are %scompressed\n", SQUASHFS_UNCOMPRESSED_INODES(sBlk.flags) ? "un" : "");
  fprintf(stderr, "Data is %scompressed\n", SQUASHFS_UNCOMPRESSED_DATA(sBlk.flags) ? "un" : "");
  if (sBlk.s_major > 1 && !SQUASHFS_NO_FRAGMENTS(sBlk.flags)) {
    fprintf(stderr, "Fragments are %scompressed\n", SQUASHFS_UNCOMPRESSED_FRAGMENTS(sBlk.flags) ? "un" : "");
  }
  fprintf(stderr, "Check data is %spresent in the filesystem\n", SQUASHFS_CHECK_DATA(sBlk.flags) ? "" : "not ");
  if (sBlk.s_major > 1) {
    fprintf(stderr, "Fragments are %spresent in the filesystem\n", SQUASHFS_NO_FRAGMENTS(sBlk.flags) ? "not " : "");
    fprintf(stderr, "Always_use_fragments option is %sspecified\n", SQUASHFS_ALWAYS_FRAGMENTS(sBlk.flags) ? "" : "not ");
  } else {
    fprintf(stderr, "Fragments are not supported by the filesystem\n");
  }
  
  if(sBlk.s_major > 1) {
    fprintf(stderr, "Duplicates are %sremoved\n", SQUASHFS_DUPLICATES(sBlk.flags) ? "" : "not ");
  } else {
    fprintf(stderr, "Duplicates are removed\n");
  }
  fprintf(stderr, "Filesystem size %.2f Kbytes (%.2f Mbytes)\n", sBlk.bytes_used / 1024.0, sBlk.bytes_used / (1024.0 * 1024.0));
  fprintf(stderr, "Block size %d\n", sBlk.block_size);
  if (sBlk.s_major > 1) {
    fprintf(stderr, "Number of fragments %d\n", sBlk.fragments);
  }
  fprintf(stderr, "Number of inodes %d\n", sBlk.inodes);
  fprintf(stderr, "Number of uids %d\n", sBlk.no_uids);
  fprintf(stderr, "Number of gids %d\n", sBlk.no_guids);
  
  TRACE("sBlk.inode_table_start 0x%llx\n", sBlk.inode_table_start);
  TRACE("sBlk.directory_table_start 0x%llx\n", sBlk.directory_table_start);
  TRACE("sBlk.uid_start 0x%llx\n", sBlk.uid_start);
  if(sBlk.s_major > 1) {
    TRACE("sBlk.fragment_table_start 0x%llx\n\n", sBlk.fragment_table_start);
  }
}

void read_uids_guids (int img_fd)
{
  if ((uid_table = malloc((sBlk.no_uids + sBlk.no_guids) * sizeof(unsigned int))) == NULL) {
    EXIT_UNSQUASH("read_uids_guids: failed to allocate uid/gid table\n");
  }

  guid_table = uid_table + sBlk.no_uids;

  if (swap_bytes) {
    unsigned int suid_table[sBlk.no_uids + sBlk.no_guids];
    
    if (read_bytes(img_fd, sBlk.uid_start, (sBlk.no_uids + sBlk.no_guids) * sizeof(unsigned int), (unsigned char *) suid_table) == FALSE) {
      EXIT_UNSQUASH("read_uids_guids: failed to read uid/gid table\n");
    }
    SQUASHFS_SWAP_INTS(uid_table, suid_table, sBlk.no_uids + sBlk.no_guids);

  } else {
    if (read_bytes(img_fd, sBlk.uid_start, (sBlk.no_uids + sBlk.no_guids) * sizeof(unsigned int), (unsigned char *) uid_table) == FALSE) {
      EXIT_UNSQUASH("read_uids_guids: failed to read uid/gid table\n");
    }
  }
}

static int add_entry(struct hash_table_entry *hash_table[], int start, int bytes)
{
  int hash = CALCULATE_HASH(start);
  struct hash_table_entry *hash_table_entry;

  if((hash_table_entry = malloc(sizeof(struct hash_table_entry))) == NULL) {
    ERROR("add_hash: out of memory in malloc\n");
    return FALSE;
  }

  hash_table_entry->start = start;
  hash_table_entry->bytes = bytes;
  hash_table_entry->next = hash_table[hash];
  hash_table[hash] = hash_table_entry;

  return TRUE;
}


static int lookup_entry(struct hash_table_entry *hash_table[], int start)
{
  int hash = CALCULATE_HASH(start);
  struct hash_table_entry *hash_table_entry;
  
  for (hash_table_entry = hash_table[hash]; hash_table_entry; hash_table_entry = hash_table_entry->next) {
    if (hash_table_entry->start == start) { return hash_table_entry->bytes; }
  }
  
  return -1;
}

static inline char * modestr (char *str, int mode)
{
  int i;

  strcpy(str, "----------");
  for(i = 0; table[i].mask != 0; i++) {
    if((mode & table[i].mask) == table[i].value) { str[table[i].position] = table[i].mode; }
  }

  return str;
}


#define TOTALCHARS  25
static int print_filename (char *pathname, struct inode *inode)
{
  char str[11], dummy[100], dummy2[100], *userstr, *groupstr;
  int padchars;
  struct passwd *user;
  struct group *group;
  struct tm *t;

  if (!unsquash_opts->do_list_extended) {
    printf("%s\n", pathname);
    return 1;
  }

  if ((user = getpwuid(inode->uid)) == NULL) {
    sprintf(dummy, "%d", inode->uid);
    userstr = dummy;
  } else
    userstr = user->pw_name;
  
  if((group = getgrgid(inode->gid)) == NULL) {
    sprintf(dummy2, "%d", inode->gid);
    groupstr = dummy2;
  } else
    groupstr = group->gr_name;
  
  printf("%s %s/%s ", modestr(str, inode->mode), userstr, groupstr);

  switch(inode->mode & S_IFMT) {
  case S_IFREG:
  case S_IFDIR:
  case S_IFSOCK:
  case S_IFIFO:
  case S_IFLNK:
    padchars = TOTALCHARS - strlen(userstr) - strlen(groupstr);
    printf("%*lld ", padchars > 0 ? padchars : 0, inode->data);
    break;
  case S_IFCHR:
  case S_IFBLK:
    padchars = TOTALCHARS - strlen(userstr) - strlen(groupstr) - 7; 
    
    printf("%*s%3d,%3d ", padchars > 0 ? padchars : 0, " ", (int) inode->data >> 8, (int) inode->data & 0xff);
    break;
  }

  t = localtime(&inode->time);
  
  printf("%d-%02d-%02d %02d:%02d %s", t->tm_year + 1900, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, pathname);
  if((inode->mode & S_IFMT) == S_IFLNK) { printf(" -> %s", inode->symlink); }
  printf("\n");
  
  return 1;
}


static int matches (struct pathnames *paths, char *name, struct pathnames **new)
{
  int i, n;

  if (paths == NULL) {
    *new = NULL;
    return TRUE;
  }

  *new = init_subdir();

  for (n = 0; n < paths->count; n++) {
    struct pathname *path = paths->path[n];
    for (i = 0; i < path->names; i++) {
      int match = (unsquash_opts->do_use_regex) ?
	regexec(path->name[i].preg, name, (size_t) 0, NULL, 0) == 0 :
	fnmatch(path->name[i].name, name, FNM_PATHNAME|FNM_PERIOD|FNM_EXTMATCH) == 0;

      if (match && path->name[i].paths == NULL) {
	/* match on a leaf component, any subdirectories will
	 * implicitly match, therefore return an empty new search set */
	goto empty_set;
      }

      if (match) {
	/* match on a non-leaf component, add any subdirectories to
	 * the new set of subdirectories to scan for this name */
	*new = add_subdir(*new, path->name[i].paths);
      }
    }
  }

  if ((*new)->count == 0) {
    /* no matching names found, delete empty search set, and return
     * FALSE */
    free_subdir(*new);
    *new = NULL;
    return FALSE;
  }

  /* one or more matches with sub-directories found (no leaf matches),
   * return new search set and return TRUE */
  return TRUE;

empty_set:
  /* found matching leaf exclude, return empty search set and return TRUE */
  free_subdir(*new);
  *new = NULL;
  return TRUE;
}


static void uncompress_inode_table (int img_fd, off_t start, off_t end)
{
  int size = 0; int bytes = 0; int res;

  while (start < end) {

    if ((size - bytes < SQUASHFS_METADATA_SIZE) &&
	((inode_table = realloc(inode_table, size += SQUASHFS_METADATA_SIZE)) == NULL)) {
      EXIT_UNSQUASH("uncompress_inode_table: out of memory in realloc\n");
    }
    
    TRACE("uncompress_inode_table: reading block 0x%llx\n", start);
    add_entry(inode_table_hash, start, bytes);
    if ((res = read_block(img_fd, start, &start, inode_table + bytes)) == 0) {
      free(inode_table);
      EXIT_UNSQUASH("uncompress_inode_table: failed to read block\n");
    }
    bytes += res;
  }
}

void do_uncompress_inode_table (int img_fd)
{
  uncompress_inode_table(img_fd, sBlk.inode_table_start, sBlk.directory_table_start);
}


static void uncompress_directory_table (int img_fd, off_t start, off_t end)
{
  int bytes = 0; int size = 0; int res;

  while (start < end) {
    if(size - bytes < SQUASHFS_METADATA_SIZE && \
       (directory_table = realloc(directory_table, size += SQUASHFS_METADATA_SIZE)) == NULL) {
    
      EXIT_UNSQUASH("uncompress_directory_table: out of memory in realloc\n");
    }
    TRACE("uncompress_directory_table: reading block 0x%llx\n", start);
    add_entry(directory_table_hash, start, bytes);
    if ((res = read_block(img_fd, start, &start, directory_table + bytes)) == 0) {
      EXIT_UNSQUASH("uncompress_directory_table: failed to read block\n");
    }
    bytes += res;
  }
}

void do_uncompress_directory_table (int img_fd)
{
  uncompress_directory_table(img_fd, sBlk.directory_table_start, sBlk.fragment_table_start);
}

int do_dir_scan (int img_fd, char *parent_name, struct pathnames *paths)
{
  return dir_scan(img_fd, parent_name, SQUASHFS_INODE_BLK(sBlk.root_inode), SQUASHFS_INODE_OFFSET(sBlk.root_inode), paths);
}

static int dir_scan (int img_fd, char *parent_name, unsigned int start_block, unsigned int offset, struct pathnames *paths)
{
  struct dir *dir = s_ops.squashfs_opendir(parent_name, start_block, offset);
  unsigned int type;
  char *name, pathname[1024];
  struct pathnames *new;

  if (dir == NULL) {
    ERROR("dir_scan: Failed to read directory %s (%x:%x)\n", parent_name, start_block, offset);
    return FALSE;
  }

  if (unsquash_opts->do_extract && mkdir(parent_name, (mode_t) dir->mode) == -1 && (!unsquash_opts->do_force_extract || errno != EEXIST)) {
    ERROR("dir_scan: failed to open directory %s, because %s\n", parent_name, strerror(errno));
    return FALSE;
  }

  while (squashfs_readdir(dir, &name, &start_block, &offset, &type)) {
    struct inode *i;
    
    TRACE("dir_scan: name %s, start_block %d, offset %d, type %d\n", name, start_block, offset, type);

    if (!matches(paths, name, &new)) { continue; }

    strcat(strcat(strcpy(pathname, parent_name), "/"), name);

    if (type == SQUASHFS_DIR_TYPE) {

      dir_scan(img_fd, pathname, start_block, offset, new);

    } else if (new == NULL) {

      if ((i = s_ops.read_inode(start_block, offset)) == NULL) {
	ERROR("failed to read header\n");
	continue;
      }
      
      if (unsquash_opts->do_list_files) { print_filename(pathname, i); }
      if (unsquash_opts->do_extract) { create_inode(img_fd, pathname, i); }
    }

    free_subdir(new);
  }

  if (unsquash_opts->do_extract) { set_attributes(parent_name, dir->mode, dir->uid, dir->guid, dir->mtime, unsquash_opts->do_force_extract); }

  squashfs_closedir(dir);
  dir_count ++;

  return TRUE;
}



#ifdef USE_MMAP

/**
 * Attempts to parse a SquashFS superblock.
 * @param img_ptr a pointer to the in-memory SquashFS image
 * @return the size of the block if we successfully parsed a superblock (possibly swapping bytes)
 *         -1 if we couldn't find a SquashFS superblock
 *         -2 if we found an unsupported superblock
 */
ssize_t squashfs_parse_superblock (const unsigned char * img_ptr)
{
  squashfs_super_block tmp_blk;

  /* Look for a valid SQUASHFS superblock */
  memcpy((unsigned char *)&tmp_blk, img_ptr, sizeof(squashfs_super_block));
  if (tmp_blk.s_magic == SQUASHFS_MAGIC) {
    swap_bytes = 0;
    memcpy((unsigned char *)&sBlk, (unsigned char *)&tmp_blk, sizeof(squashfs_super_block));
  } else if (tmp_blk.s_magic == SQUASHFS_MAGIC_SWAP) {
    swap_bytes = 1;
    SQUASHFS_SWAP_SUPER_BLOCK(&sBlk, &tmp_blk);
  } else {
    return -1;
  }

  /* Check major & minor versions */
  if (sBlk.s_major == 1 || sBlk.s_major == 2) {
    sBlk.bytes_used = sBlk.bytes_used_2;
    sBlk.uid_start = sBlk.uid_start_2;
    sBlk.guid_start = sBlk.guid_start_2;
    sBlk.inode_table_start = sBlk.inode_table_start_2;
    sBlk.directory_table_start = sBlk.directory_table_start_2;
		
    if (sBlk.s_major == 1) {
      sBlk.block_size = sBlk.block_size_1;
      sBlk.fragment_table_start = sBlk.uid_start;
      s_ops.squashfs_opendir = squashfs_opendir_2;
      s_ops.read_fragment_table = read_fragment_table_1;
      s_ops.read_block_list = read_block_list_1;
      s_ops.read_inode = read_inode_1;
    } else {
      sBlk.fragment_table_start = sBlk.fragment_table_start_2;
      s_ops.squashfs_opendir = squashfs_opendir_2;
      s_ops.read_fragment = read_fragment_2;
      s_ops.read_fragment_table = read_fragment_table_2;
      s_ops.read_block_list = read_block_list;
      s_ops.read_inode = read_inode_2;
    }
  } else if (sBlk.s_major == 3 && sBlk.s_minor <= 1) {
    s_ops.squashfs_opendir = squashfs_opendir;
    s_ops.read_fragment = read_fragment;
    s_ops.read_fragment_table = read_fragment_table;
    s_ops.read_block_list = read_block_list;
    s_ops.read_inode = read_inode;
  } else {
    ERROR("Found unsupported superblock version (%d:%d), ", sBlk.s_major, sBlk.s_minor);
    return -2;
  }

  return (int)swap_bytes;
}


static inline ssize_t parse_id_tables (const unsigned char * img_ptr)
{
  size_t uid_table_len; size_t gid_table_len; ssize_t err;
  unsigned int * tmp_uid_table; unsigned int * tmp_gid_table;

  uid_table_len = sBlk.no_uids * sizeof(unsigned int);
  gid_table_len = sBlk.no_guids * sizeof(unsigned int);

  tmp_uid_table = malloc(uid_table_len);
  tmp_gid_table = malloc(gid_table_len);

  if (tmp_uid_table && tmp_gid_table) {
    memcpy(tmp_uid_table, img_ptr, uid_table_len); img_ptr += uid_table_len;
    memcpy(tmp_gid_table, img_ptr, gid_table_len); img_ptr += gid_table_len;
  
    if ((uid_table = malloc(uid_table_len)) && \
	(guid_table = malloc(gid_table_len)) ) {

      if (swap_bytes) {
	SQUASHFS_SWAP_INTS(uid_table, tmp_uid_table, uid_table_len);
	SQUASHFS_SWAP_INTS(guid_table, tmp_gid_table, gid_table_len);
      } else {
	memcpy(uid_table, tmp_uid_table, uid_table_len);
	memcpy(guid_table, tmp_gid_table, gid_table_len);
      }
      err = uid_table_len + gid_table_len;

    } else {
      ERROR("Failed to read UID/GID table.\n"); err = -1;
    }

  } else {
    ERROR("Failed to read UID/GID table.\n"); err = -1;
  }

  free(tmp_uid_table); free(tmp_gid_table);
  return err;
}

static inline ssize_t parse_fragment_table (const unsigned char * img_ptr)
{
  return 0;
}

/**
 * Parses the SquashFS metadata.
 * @param img_ptr a pointer to the image metadata
 * @return the pointer offset if everything went OK
 *         a negative error code otherwise
 */
ssize_t squashfs_parse_metadata (const unsigned char * img_ptr)
{
  ssize_t err;

  if ((err = parse_id_tables(img_ptr)) > 0) {  /* First read the UID/GID tables */
    img_ptr += err;

    /* Then read the fragment table */

  }

  return 0;
  
}

#endif
