/*
 * Utility functions the image tools.
 *
 * Copyright (c) Simon Vallet 2008 <svallet@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * utils.c
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <zlib.h>

#include <bcm_img.h>
#include <utils.h>

/***** Options *****/

/** bcm_kimg */
struct kimg_opts * kimg_opts = NULL;

static inline void kimg_usage (void)
{
  fprintf(stderr, "Usage: bcm_kimg -c|-x [-v] [-d <dir>] [-r <raw kernel>] [-e <ELF kernel>] [-z <LZMA kernel>] [<image_file>]\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -c:\tcreate a kernel image\n");
  fprintf(stderr, "  -x:\textract components from a kernel image\n");
  fprintf(stderr, "  -v:\tverbose output\n");
  fprintf(stderr, "  -B <addr>:\tspecify the kernel load base, defaults to 0x%.8x\n", _KIMG_DFLT_LOAD_ADDR);
  fprintf(stderr, "  -d <dir>:\textract to/build from <dir>\n");
  fprintf(stderr, "  -r <file>:\traw kernel file name, defaults to %s\n", _DFLT_RK_FILE);
  fprintf(stderr, "  -e <file>:\tELF kernel file name, defaults to %s\n", _DFLT_EK_FILE);
  fprintf(stderr, "  -z <file>:\tLZMA kernel file name, defaults to %s\n", _DFLT_ZK_FILE);
  fprintf(stderr, "  <image_file>:\tBroadcom kernel image, defaults to %s\n", _DFLT_KIMG_FILE);
}

/* Returns the offset of the last option */
int parse_kimg_opts (int arg_count, char ** args)
{
  int optbuf;
  
  if ((kimg_opts = (struct kimg_opts *)malloc(sizeof(struct kimg_opts)))) {

    /* Handle defaults */
    kimg_opts->todo = _KIMG_DO_NOTHING;
    kimg_opts->be_verbose = 0;
    kimg_opts->dir = NULL;
    kimg_opts->raw_kernel = _DFLT_RK_FILE;
    kimg_opts->elf_kernel = _DFLT_EK_FILE;
    kimg_opts->lz_kernel = _DFLT_ZK_FILE;
    kimg_opts->kernel_img = _DFLT_KIMG_FILE;
    kimg_opts->load_base = _KIMG_DFLT_LOAD_ADDR;

    /* Opts */
    while ((optbuf=getopt(arg_count, args, "d:r:e:z:B:cxvh")) != -1) {
      switch(optbuf) {
      case 'd':
	kimg_opts->dir = optarg;
        break;
      case 'r':
	kimg_opts->raw_kernel = optarg;
	break;
      case 'e':
	kimg_opts->elf_kernel = optarg;
        break;
      case 'z':
	kimg_opts->lz_kernel = optarg;
        break;
      case 'c':
	if (kimg_opts->todo == _KIMG_DO_EXTRACT) {
	  fprintf(stderr, "Invocation error: '-c' and '-x' are mutually exclusive.\n");
	  return -1;
	}
	kimg_opts->todo = _KIMG_DO_CREATE;
        break;
      case 'x':
	if (kimg_opts->todo == _KIMG_DO_CREATE) {
	  fprintf(stderr, "Invocation error: '-c' and '-x' are mutually exclusive.\n");
	  return -1;
	}
	kimg_opts->todo = _KIMG_DO_EXTRACT;
	break;
      case 'B':
	errno = 0;
	kimg_opts->load_base = strtoul(optarg, NULL, 16);
	if (errno) {
	  fprintf(stderr, "Load base has an invalid format (should be hexdecimal)\n");
	  return -1;
	}
	break;
      case 'v':
	kimg_opts->be_verbose = 1;
	break;
      case 'h':
      default:
	kimg_usage();
        return -1;
        break;
      }
    }
    
    if (kimg_opts->todo == _KIMG_DO_NOTHING) {
      fprintf(stderr, "At least one of '-c' or '-x' has to be indicated ('-h' for usage).\n");
      return -1;
    }

    if (arg_count - optind >= 1) {
      kimg_opts->kernel_img = args[optind++];
    }
  }

  return optind;
}


/** LZMA unsquash **/

struct unsquash_opts * unsquash_opts = NULL;

static inline void lzma_unsquash_version (void)
{
  fprintf(stderr, "lzma_unsquash version 0.1 (2008/03)\n");
  fprintf(stderr, "Based on unsquashfs version 1.5 (2007/10/31)\n");
  fprintf(stderr, "Copyright (C) 2007 Phillip Lougher <phillip@lougher.demon.co.uk>\n");
  fprintf(stderr, "Copyright (C) 2008 Simon Vallet <svallet@gmail.com>\n\n");
  fprintf(stderr, "This program is free software; you can redistribute it and/or\n");
  fprintf(stderr, "modify it under the terms of the GNU General Public License\n");
  fprintf(stderr, "as published by the Free Software Foundation; either version 2,\n");
  fprintf(stderr, "or (at your option) any later version.\n\n");
  fprintf(stderr, "This program is distributed in the hope that it will be useful,\n");
  fprintf(stderr, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  fprintf(stderr, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  fprintf(stderr, "GNU General Public License for more details.\n");
}

static inline void lzma_unsquash_usage (const char * progname)
{
  fprintf(stderr, "Usage: %s [options] filesystem [directories or files to extract]\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -V\t\tprint version, licence and copyright information\n");
  fprintf(stderr, "  -v\t\tlist files as they are unsquashed\n");
  fprintf(stderr, "  -l\t\tlist filesystem, but don't unsquash\n");
  fprintf(stderr, "  -x\t\tprint extended file attributes when listing (like ls -l output)\n");
  fprintf(stderr, "  -d <dir>\tunsquash to <dir>, default \"squashfs-root\"\n");
  fprintf(stderr, "  -f\t\tif the output file already exists then overwrite it\n");
  fprintf(stderr, "  -s\t\tdisplay filesystem superblock information\n");
  /*
			ERROR("\t-e[f] <extract file>\tlist of directories or files to extract.\n\t\t\t\tOne per line\n");
			ERROR("\t-r[egex]\t\ttreat extract names as POSIX regular expressions\n\t\t\t\trather than use the default shell wildcard\n\t\t\t\texpansion (globbing)\n");
  */
}

/* Returns the offset of the last option */
int parse_unsquash_opts (int arg_count, char ** args)
{
  int optbuf;
  
  if ((unsquash_opts = (struct unsquash_opts *)malloc(sizeof(struct unsquash_opts)))) {

    /* Handle defaults */
    unsquash_opts->show_superblock_only = 0;
    unsquash_opts->do_list_files = 0;
    unsquash_opts->do_list_extended = 0;
    unsquash_opts->do_extract = 1; /* TODO */
    unsquash_opts->do_force_extract = 0;
    unsquash_opts->do_use_regex = 0;
    unsquash_opts->be_quiet = 0;
    unsquash_opts->output_dir = NULL;

    /* Unsquash_Opts */
    while ((optbuf=getopt(arg_count, args, "d:hvVlxfs")) != -1) {
      switch(optbuf) {
      case 'f':
        unsquash_opts->do_force_extract = 1;
        break;
      case 'l':
	unsquash_opts->do_list_files = 1;
	unsquash_opts->do_extract = 0;
	break;
      case 's':
        unsquash_opts->show_superblock_only = 1;
        break;
      case 'x':
        unsquash_opts->do_list_extended = 1;
        break;
      case 'd':
        unsquash_opts->output_dir=optarg;
        break;
      case 'v':
	unsquash_opts->do_list_files = 1;
	break;
      case 'V':
	lzma_unsquash_version();
	return -1;
	break;
      case 'h':
      default:
	lzma_unsquash_usage(args[0]);
        return -1;
        break;
      }
    }
    
    if (arg_count - optind < 1) {
      lzma_unsquash_usage(args[0]);
      return -1;
    }
  }

  return optind;
}

/** FW extraction **/

struct extract_opts * extract_opts = NULL;

static inline void fw_extract_usage (void)
{
  fprintf(stderr, "Usage: bcm_fw_extract [-bfhkaxv] [-d <dir>] <image_file>\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "\t-b: dump the CFE bootloader\n");
  fprintf(stderr, "\t-f: dump the root FS\n");
  fprintf(stderr, "\t-f: dump the Broadcom header\n");
  fprintf(stderr, "\t-k: dump the kernel\n");
  fprintf(stderr, "\t-a: dump everything\n");
  fprintf(stderr, "\t-x: extract dumped sections (rootfs and/or kernel)\n");
  fprintf(stderr, "\t-v: verbose header information\n");
  fprintf(stderr, "\t-d <dir>: dump files into <dir>\n");
}

int parse_fw_extract_opts (int arg_count, char ** args)
{
  int optbuf;

  if ((extract_opts = (struct extract_opts *)malloc(sizeof(struct extract_opts)))) {

    /* Handle defaults */
    extract_opts->output_dir = NULL;
    extract_opts->dump_header = 0;
    extract_opts->dump_cfe = 0;
    extract_opts->dump_rootfs = 0;
    extract_opts->dump_kernel = 0;
    extract_opts->be_verbose = 0;
    extract_opts->do_extract = 0;

    /* Extract_Opts */
    while ((optbuf=getopt(arg_count, args, "d:bfhkaxv")) != -1) {
      switch (optbuf) {
      case 'b':
        extract_opts->dump_cfe = 1;
        break;
      case 'f':
	extract_opts->dump_rootfs = 1;
	break;
      case 'h':
	extract_opts->dump_header = 1;
	break;
      case 'k':
        extract_opts->dump_kernel = 1;
        break;
      case 'a':
        extract_opts->dump_cfe = 1;
	extract_opts->dump_header = 1;
        extract_opts->dump_rootfs = 1;
        extract_opts->dump_kernel = 1;
        break;
      case 'x':
	extract_opts->do_extract = 1;
	break;
      case 'd':
        extract_opts->output_dir=optarg;
        break;
      case 'v':
	extract_opts->be_verbose=1;
	break;
      default:
	fw_extract_usage();
        return -1;
        break;
      }
    }
    
    if (arg_count - optind != 1) {
      fw_extract_usage();
      return -1;
    }
  }

  return optind;
}

/** FW creation **/
struct create_opts * create_opts = NULL;

static inline void fw_create_usage (void)
{
  fprintf(stderr, "Usage: bcm_fw_create [-abfkr] [-d <dir>] [-B <board_id>] [-C <chip_id>] <image_file>\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "\t-b: include a bootloader image (default: %s)\n", _BOOTLDR_FILE);
  fprintf(stderr, "\t-f: include a filesystem image (default: %s)\n", _ROOTFS_FILE);
  fprintf(stderr, "\t-k: include a kernel image (default %s)\n", _DFLT_KIMG_FILE);
  fprintf(stderr, "\t-a: include everything\n");
  fprintf(stderr, "\t-r: create a \"raw\" image\n");
  fprintf(stderr, "\t-d <dir>: image files are located in <dir>\n");
  fprintf(stderr, "\t-B <board_id>: specify the board id (default: %s)\n", _DFLT_BOARD_ID);
  fprintf(stderr, "\t-C <chip_id>: specify the chip id (default: %s)\n", _DFLT_CHIP_ID);
}


int parse_fw_create_opts (int arg_count, char ** args)
{
  int optbuf;

  if ((create_opts = (struct create_opts *)malloc(sizeof(struct create_opts)))) {

    /* Handle defaults */
    create_opts->input_dir = NULL;
    create_opts->board_id = _DFLT_BOARD_ID;
    create_opts->chip_id = _DFLT_CHIP_ID;
    create_opts->include_bootldr = 0;
    create_opts->include_rootfs = 0;
    create_opts->include_kernel = 0;
    create_opts->include_bcm_hdr = 1;

    /* 0pts */
    while ((optbuf=getopt(arg_count, args, "d:B:C:abkfr")) != -1) {
      switch (optbuf) {
      case 'a':
	create_opts->include_bootldr = 1;
	create_opts->include_kernel = 1;
	create_opts->include_rootfs = 1;
	break;
      case 'b':
	create_opts->include_bootldr = 1;
	break;
      case 'B':
	create_opts->board_id = optarg;
	break;
      case 'C':
	create_opts->chip_id = optarg;
	break;
      case 'd':
	create_opts->input_dir = optarg;
	break;
      case 'k':
	create_opts->include_kernel = 1;
	break;
      case 'f':
	create_opts->include_rootfs = 1;
	break;
      case 'r':
	create_opts->include_bcm_hdr = 0;
	break;
      default:
	fw_create_usage();
	return -1;
	break;
      }
    }

    if (arg_count - optind != 1) {
      fw_create_usage();
      return -1;
    }
  }

  return optind;
}

/***** Misc utils *****/

/* Returns the CRC used in Broadcom images */
u_int32_t bcm_crc32 (const unsigned char * ptr, size_t len)
{
  u_int32_t crc;

  crc = crc32(0UL, NULL, 0);	
  crc = crc32(crc, ptr, len);

  return ~crc;
}

/* Returns 1 if host is big endian, 0 otherwise */
int is_big_endian (void)
{
  unsigned int check = 0xdeadbeef;

  return (((unsigned char *)&check)[0] == 0xde);
}

u_int32_t swap_bytes_32 (const u_int32_t in)
{
  u_int32_t out;

  ((unsigned char *)&out)[0] = ((unsigned char *)&in)[3];
  ((unsigned char *)&out)[1] = ((unsigned char *)&in)[2];
  ((unsigned char *)&out)[2] = ((unsigned char *)&in)[1];
  ((unsigned char *)&out)[3] = ((unsigned char *)&in)[0];

  return out;
}

/* Dumps a hex representation of the given string */
void hex_dump (const unsigned char * input, size_t sz)
{
  size_t i;

  for (i=0; i<sz; i++) {
    /* fprintf(stderr, "0x%.2x ", input[i]); */
    if (i%16 == 0) { fprintf(stderr, "\n\t"); }
    fprintf(stderr, "%.2x ", input[i]);
  }
}

